/*
 *  The contents of this file are subject to the Initial
 *  Developer's Public License Version 1.0 (the "License");
 *  you may not use this file except in compliance with the
 *  License. You may obtain a copy of the License at
 *  http://www.ibphoenix.com/main.nfs?a=ibphoenix&page=ibp_idpl.
 *
 *  Software distributed under the License is distributed AS IS,
 *  WITHOUT WARRANTY OF ANY KIND, either express or implied.
 *  See the License for the specific language governing rights
 *  and limitations under the License.
 *
 *  The Original Code was created by Dmitry Yemanov
 *  for the Firebird Open Source RDBMS project.
 *
 *  Copyright (c) 2013 Dmitry Yemanov <dimitr@firebirdsql.org>
 *  and all contributors signed below.
 *
 *  All Rights Reserved.
 *  Contributor(s): ______________________________________.
 */

#include "firebird.h"
#include "ids.h"
#include "../common/classes/alloc.h"
#include "../common/classes/auto.h"
#include "../common/classes/ClumpletWriter.h"
#include "../common/isc_proto.h"
#include "../dsql/sqlda_pub.h"
#include "../jrd/jrd.h"
#include "../jrd/blb.h"
#include "../jrd/ods.h"
#include "../jrd/req.h"
#include "../jrd/ini.h"
#include "../jrd/sqz.h"
#include "../jrd/ibase.h"
#include "../jrd/btr_proto.h"
#include "../jrd/cch_proto.h"
#include "../jrd/cmp_proto.h"
#include "../jrd/dpm_proto.h"
#include "../jrd/evl_proto.h"
#include "../jrd/idx_proto.h"
#include "../jrd/inf_proto.h"
#include "../jrd/ini_proto.h"
#include "../jrd/jrd_proto.h"
#include "../jrd/lck_proto.h"
#include "../jrd/met_proto.h"
#include "../jrd/mov_proto.h"
#include "../jrd/pag_proto.h"
#include "../jrd/rlck_proto.h"
#include "../jrd/tra_proto.h"
#include "../jrd/vio_proto.h"
#include "../jrd/evl_string.h"
#include "../jrd/SimilarToMatcher.h"
#include "../jrd/trace/TraceManager.h"
#include "../jrd/trace/TraceJrdHelpers.h"

#include "Replication.h"
#include "ReplicationLog.h"
#include "ReplicationUtils.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif
#include <stdlib.h>

// Debugging macros
//#define REPL_LOG
//#define REPL_NOP

//#define LOG_WARNINGS

using namespace Firebird;
using namespace Ods;
using namespace Jrd;
using namespace Replication;
using namespace ReplicationUtils;

const size_t GUID_STR_LENGTH = GUID_BUFF_SIZE - 1;
const size_t INIT_PACKET_SIZE = sizeof(UCHAR) + sizeof(ULONG) * 3 + GUID_STR_LENGTH;
const size_t START_TXN_PACKET_SIZE = sizeof(UCHAR) + sizeof(AttNumber) + sizeof(TraNumber);
const size_t PREPARE_TXN_PACKET_SIZE = sizeof(UCHAR) + sizeof(TraNumber);

const size_t MAX_BG_WRITER_LAG = 10 * 1024 * 1024;	// 10 MB

const char* SYNC_WARNING = "Slave is to be detached after error. Replica may become inconsistent at this point.";

namespace
{
	struct NoKeyTable
	{
		USHORT rel_id;
		USHORT rel_fields[8];
	};

	NoKeyTable NO_KEY_TABLES[] = {
		{ rel_segments, { f_seg_name, f_seg_field, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT } },
		{ rel_args, { f_arg_fun_name, f_arg_pos, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT } },
		{ rel_ccon, { f_ccon_cname, f_ccon_tname, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT } },
		{ rel_vrel, { f_vrl_vname, f_vrl_context, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT } },
		{ rel_msgs, { f_msg_trigger, f_msg_number, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT } },
		{ rel_dims, { f_dims_fname, f_dims_dim, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT } },
		{ rel_files, { f_file_name, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT, MAX_USHORT } },
		{ rel_priv, { f_prv_user, f_prv_u_type, f_prv_o_type, f_prv_priv, f_prv_grant, f_prv_grantor, f_prv_rname, f_prv_fname } }
	};


	class ReplicationManager : public RefCounted, public GlobalStorage, public ErrorLogger
	{
	public:
		struct QueueItem
		{
			explicit QueueItem(MemoryPool& pool)
				: status(pool), buffer(pool),
				  last_mod(TimeStamp::getCurrentTimeStamp()),
				  wf_count(0), wf_time(0), wf_transfer(0),
				  bg_count(0), bg_time(0), bg_transfer(0)
			{}

			FbLocalStatus status;
			UCharBuffer buffer;
			Mutex mutex;
			Firebird::TimeStamp last_mod;
			ULONG wf_count;
			SINT64 wf_time;
			FB_UINT64 wf_transfer;
			ULONG bg_count;
			SINT64 bg_time;
			FB_UINT64 bg_transfer;
		};

		typedef GenericMap<Pair<Left<string, ReplicationManager*> > > DbReplMgrMap;
		typedef GenericMap<Pair<NonPooled<FB_API_HANDLE, QueueItem*> > > FlushQueue;

	public:
		static ReplicationManager* create(IReplicationHelper* helper,
										  ReplicationConfig* config)
		{
			PathName database(helper->getDatabaseName());

			const Guid* const guid = helper->getDatabaseGuid();

			char buff[GUID_BUFF_SIZE];
			GuidToString(buff, guid);
			const string guidStr(buff);
			fb_assert(guidStr.length() == GUID_STR_LENGTH);

			MutexLockGuard guard(g_mapMutex, FB_FUNCTION);

			ReplicationManager* mgr = NULL;
			if (!g_map->get(guidStr, mgr))
			{
				mgr = FB_NEW ReplicationManager(database, guidStr, *guid, config);

				if (g_map->put(guidStr, mgr))
					fb_assert(false);
			}

			fb_assert(mgr);

			mgr->addRef();
			return mgr;
		}

		static void destroy(ReplicationManager* mgr)
		{
			if (mgr)
			{
				const Firebird::string dbId = mgr->m_dbId;

				Firebird::MutexLockGuard guard(g_mapMutex, FB_FUNCTION);

				if (!mgr->release())
				{
					if (!g_map->remove(dbId))
						fb_assert(false);
				}
			}
		}

		~ReplicationManager()
		{
			m_shutdown = true;

			m_workingSemaphore.release();
			m_cleanupSemaphore.enter();

			WriteLockGuard guard(m_lock, FB_FUNCTION);

			FlushQueue::Accessor accessor(&m_queue);
			if (accessor.getFirst())
			{
				do {
					delete accessor.current()->second;
					accessor.current()->second = NULL;
				} while (accessor.getNext());
			}

			delete m_log;
		}

		void init(FB_API_HANDLE handle)
		{
			++m_waiters;
			WriteLockGuard guard(m_lock, FB_FUNCTION);
			--m_waiters;

			QueueItem* item = NULL;
			if (!m_queue.get(handle, item))
			{
				MemoryPool& pool = getPool();
				item = FB_NEW_POOL(pool) QueueItem(pool);
				m_queue.put(handle, item);
			}
		}

		void cleanup(FB_API_HANDLE handle)
		{
			++m_waiters;
			WriteLockGuard guard(m_lock, FB_FUNCTION);
			--m_waiters;

			QueueItem* item = NULL;
			if (m_queue.get(handle, item))
			{
				delete item;
				m_queue.remove(handle);
			}
		}

		void replicate(FB_API_HANDLE handle, QueueItem* item, bool force)
		{
			try
			{
				if (item->status[1] && !force)
					status_exception::raise(&item->status);

				if (item->buffer.hasData())
				{
					ISC_STATUS_ARRAY temp_status;

					const SINT64 start_clock = fb_utils::query_performance_counter();

					try
					{
						if (handle)
						{
							fb_replicate(temp_status, &handle,
										 (ULONG) item->buffer.getCount(),
										 item->buffer.begin());

							if (temp_status[1])
								status_exception::raise(temp_status);
						}
						else
						{
							m_log->write((ULONG) item->buffer.getCount(),
										 item->buffer.begin());
						}
					}
					catch (const Exception& ex)
					{
						ex.stuffException(&item->status);
					}

					const SINT64 result_time =
						(fb_utils::query_performance_counter() - start_clock) * 1000 /
						fb_utils::query_performance_frequency();

					if (!item->status[1])
					{
						item->wf_count++;
						item->wf_time += result_time;
						item->wf_transfer += (FB_UINT64) item->buffer.getCount();
						item->last_mod = TimeStamp::getCurrentTimeStamp();
					}

					item->buffer.clear();
				}

				if (item->status[1] && !force)
					status_exception::raise(&item->status);
			}
			catch (const Exception&)
			{
				item->status->init();
				throw;
			}
		}

		void flush(FB_API_HANDLE handle, ULONG length, const UCHAR* ptr, bool sync, bool force)
		{
			fb_assert(length && ptr);

			// For asynchronous replication, don't write initialization packet to the log.

			if (!handle && *ptr == eInitialize)
			{
				length -= INIT_PACKET_SIZE;
				ptr += INIT_PACKET_SIZE;
			}

			ReadLockGuard guard(m_lock, FB_FUNCTION);

			QueueItem* item = NULL;
			if (m_queue.get(handle, item))
			{
				MutexLockGuard mutexGuard(item->mutex, FB_FUNCTION);

				item->buffer.add(ptr, length);

				// If the background thread is lagging too far behind,
				// replicate packets synchronously rather than relying
				// on the background thread to catch up any time soon
				if (item->buffer.getCount() > MAX_BG_WRITER_LAG)
					sync = true;
			}

			if (item)
			{
				if (sync)
				{
					thread_db* const tdbb = JRD_get_thread_data();

					EngineCheckout cout(tdbb, FB_FUNCTION);
					MutexLockGuard mutexGuard(item->mutex, FB_FUNCTION);

					replicate(handle, item, force);
				}
				else if (!m_signalled)
				{
					m_signalled = true;
					m_workingSemaphore.release();
				}
			}
		}

		const QueueItem* getItem(FB_API_HANDLE handle)
		{
			QueueItem* item = NULL;
			m_queue.get(handle, item);
			return item;
		}

	private:
		ReplicationManager(const PathName& database,
						   const string& dbId,
						   const Guid& guid,
						   ReplicationConfig* config)
			: ErrorLogger(getPool(), database, MASTER),
			  m_dbId(getPool(), dbId), m_queue(getPool()),
			  m_shutdown(false), m_signalled(false)
		{
			m_log = config->isLoggingActive() ?
				FB_NEW_POOL(getPool()) ReplicationLog(getPool(), config, database, guid, true) : NULL;

			Thread::start(writer_thread, this, THREAD_high, 0);
			m_startupSemaphore.enter();
		}

		void bgWriter()
		{
			m_startupSemaphore.release();

			while (!m_shutdown)
			{
				bool restartLoop = false;

				ReadLockGuard guard(m_lock, FB_FUNCTION);

				FlushQueue::Accessor accessor(&m_queue);
				if (accessor.getFirst())
				{
					do {
						if (m_shutdown)
							break;

						if (m_waiters.value())
						{
							restartLoop = true;
							break;
						}

						FB_API_HANDLE handle = accessor.current()->first;
						QueueItem* const item = accessor.current()->second;

						if (!item->mutex.tryEnter(FB_FUNCTION))
						{
							restartLoop = true;
							continue;
						}

						if (item->status[1] || item->buffer.isEmpty())
						{
							item->mutex.leave();
							continue;
						}

						ISC_STATUS_ARRAY temp_status;

						const SINT64 m_start_clock = fb_utils::query_performance_counter();

						try
						{
							if (handle)
							{
								fb_replicate(temp_status, &handle,
											 (ULONG) item->buffer.getCount(),
											 item->buffer.begin());

								if (temp_status[1])
									status_exception::raise(temp_status);
							}
							else
							{
								m_log->write((ULONG) item->buffer.getCount(),
											 item->buffer.begin());
							}
						}
						catch (const Exception& ex)
						{
							ex.stuffException(&item->status);
						}

						SINT64 m_clock = (fb_utils::query_performance_counter() - m_start_clock) * 1000 /
							fb_utils::query_performance_frequency();

						item->bg_count++;
						item->bg_time += m_clock;
						item->bg_transfer += (SINT64)item->buffer.getCount();
						item->last_mod = TimeStamp::getCurrentTimeStamp();

						item->buffer.clear();
						item->mutex.leave();

					} while (accessor.getNext());
				}

				if (m_shutdown)
					break;

				if (restartLoop)
					continue;

				guard.release();

				m_signalled = false;
				m_workingSemaphore.tryEnter(1);
			}

			m_cleanupSemaphore.release();
		}

		static THREAD_ENTRY_DECLARE writer_thread(THREAD_ENTRY_PARAM arg)
		{
			ReplicationManager* const mgr = static_cast<ReplicationManager*>(arg);
			mgr->bgWriter();
			return 0;
		}

		Semaphore m_startupSemaphore;
		Semaphore m_cleanupSemaphore;
		Semaphore m_workingSemaphore;

		string m_dbId;
		FlushQueue m_queue;

		volatile bool m_shutdown;
		volatile bool m_signalled;
		AtomicCounter m_waiters;

		ReplicationLog* m_log;
		RWLock m_lock;

		static GlobalPtr<DbReplMgrMap> g_map;
		static GlobalPtr<Mutex> g_mapMutex;
	};

	GlobalPtr<ReplicationManager::DbReplMgrMap> ReplicationManager::g_map;
	GlobalPtr<Mutex> ReplicationManager::g_mapMutex;

} // namespace

namespace Jrd
{
	class ReplicationMaster : public IReplicationMaster,
		private PermanentStorage, public ErrorLogger
	{
		friend class GlobalManager;

		typedef HalfStaticArray<SavNumber, 16> SavepointStack;
		typedef UpcaseConverter<NullStrConverter> SimilarConverter;
		typedef SimilarToMatcher<SimilarConverter, UCHAR> SimilarMatcher;

		struct TransactionData
		{
			Writer* buffer;
			SavepointStack* stack;
			ULONG flushes;
			AttNumber session;
		};

		typedef GenericMap<Pair<NonPooled<TraNumber, TransactionData*> > > TransactionMap;
		typedef GenericMap<Pair<Left<MetaName, bool> > > TablePermissionMap;

		enum FlushReason
		{
			FLUSH_OVERFLOW,
			FLUSH_PREPARE,
			FLUSH_SYNC
		};

	public:
		bool cleanupAttachment(FbStatusVector* status, AttNumber att_id);
		bool startTransaction(FbStatusVector* status, AttNumber att_id, TraNumber txn_id);
		bool prepareTransaction(FbStatusVector* status, TraNumber txn_id);
		bool commitTransaction(FbStatusVector* status, TraNumber txn_id);
		bool rollbackTransaction(FbStatusVector* status, TraNumber txn_id);
		bool cleanupTransactions(FbStatusVector* status);
		bool startSavepoint(FbStatusVector* status, TraNumber txn_id, SavNumber sav_id);
		bool cleanupSavepoint(FbStatusVector* status, TraNumber txn_id, SavNumber sav_id, bool undo);
		bool insertRecord(FbStatusVector* status, TraNumber txn_id, const char* rel_name,
							ULONG length, const UCHAR* data);
		bool updateRecord(FbStatusVector* status, TraNumber txn_id, const char* rel_name,
							ULONG org_length, const UCHAR* org_data,
							ULONG new_length, const UCHAR* new_data);
		bool deleteRecord(FbStatusVector* status, TraNumber txn_id, const char* rel_name,
							ULONG length, const UCHAR* data);
		bool changeGenerator(FbStatusVector* status, SLONG gen_id, SINT64 value);
		bool storeBlob(FbStatusVector* status, TraNumber txn_id, ISC_QUAD blob_id,
						 ULONG length, const UCHAR* data);
		void flush(FbStatusVector* status, bool sync);
		void shutdown(FbStatusVector* status, bool drop);
		bool isTableEnabled(thread_db* tdbb, jrd_rel* const relation);
		void clearTableEnabled(const char* rel_name);
		bool isMasterPriority();

	ReplicationMaster(MemoryPool& pool, IReplicationHelper* helper,
					  ULONG dpb_length, const UCHAR* dpb, bool create_flag)
		: PermanentStorage(pool),
		  ErrorLogger(pool, helper->getDatabaseName(), MASTER),
		  m_config(ReplicationConfig::get(helper->getDatabaseName())),
		  m_manager(ReplicationManager::create(helper, m_config)),
		  m_writer(pool, m_config->bufferSize),
		  m_map(pool), m_tables(pool), m_slaves(pool), m_sessions(pool),
		  m_bytes(0), m_alerted(false),
		  m_cs(FB_NEW_POOL(pool) charset), m_tt(FB_NEW_POOL(pool) texttype)
	{
		const Guid* const guid = helper->getDatabaseGuid();
		memcpy(&m_guid, guid, sizeof(Guid));

		IntlUtil::initUtf8Charset(m_cs);

		string collAttributes("ICU-VERSION=");
		collAttributes += Jrd::UnicodeUtil::getDefaultIcuVersion();
		IntlUtil::setupIcuAttributes(m_cs, collAttributes, "", collAttributes);

		UCharBuffer collAttributesBuffer;
		collAttributesBuffer.push(reinterpret_cast<const UCHAR*>(collAttributes.c_str()),
								  collAttributes.length());

		if (!IntlUtil::initUnicodeCollation(m_tt, m_cs, "UNICODE", 0, collAttributesBuffer, ""))
			raiseError("Cannot initialize UNICODE collation");

		m_charSet = CharSet::createInstance(pool, 0, m_cs);
		m_textType = FB_NEW_POOL(pool) TextType(0, m_tt, m_charSet);

		if (m_config->includeFilter.hasData())
		{
			m_includeMatcher = FB_NEW_POOL(pool) SimilarToMatcher<UCHAR, UpcaseConverter<> >(
				pool, m_textType,
				(const UCHAR*) m_config->includeFilter.c_str(),
				m_config->includeFilter.length(),
				'\\', true);
		}

		if (m_config->excludeFilter.hasData())
		{
			m_excludeMatcher = FB_NEW_POOL(pool) SimilarToMatcher<UCHAR, UpcaseConverter<> >(
				pool, m_textType,
				(const UCHAR*) m_config->excludeFilter.c_str(),
				m_config->excludeFilter.length(),
				'\\', true);
		}

		initialize(dpb_length, dpb, create_flag);
	}

	~ReplicationMaster()
	{
		if (m_tt && m_tt->texttype_fn_destroy)
			m_tt->texttype_fn_destroy(m_tt);

		TransactionMap::Accessor accessor1(&m_map);
		if (accessor1.getFirst())
		{
			do {
				delete accessor1.current()->second->buffer;
				delete accessor1.current()->second->stack;
				delete accessor1.current()->second;
			} while (accessor1.getNext());
		}

		ReplicationManager::destroy(m_manager);
	}

	private:
		RefPtr<ReplicationConfig> m_config;

		Guid m_guid;
		ReplicationManager* m_manager;
		Writer m_writer;
		TransactionMap m_map;
		TablePermissionMap m_tables;
		Array<FB_API_HANDLE> m_slaves;
		SortedArray<AttNumber> m_sessions;
		FB_UINT64 m_bytes;
		Mutex m_mutex;
		bool m_alerted;

		charset* const m_cs;
		texttype* const m_tt;
		AutoPtr<CharSet> m_charSet;
		AutoPtr<TextType> m_textType;
		AutoPtr<SimilarToMatcher<UCHAR, UpcaseConverter<> > > m_includeMatcher;
		AutoPtr<SimilarToMatcher<UCHAR, UpcaseConverter<> > > m_excludeMatcher;

		void initialize(ULONG org_dpb_length, const UCHAR* org_dpb, bool create_flag)
		{
			const unsigned replicaCount = (unsigned) m_config->replicas.getCount();
			unsigned errorCount = 0;

			if (m_config->isLoggingActive())
				m_manager->init(0);

			if (!replicaCount && m_config->isLoggingActive())
			{
				m_slaves.add(0);
				return;
			}

			char buff[GUID_BUFF_SIZE];
			GuidToString(buff, &m_guid);
			const string guidStr(buff);
			fb_assert(guidStr.length() == GUID_STR_LENGTH);

			for (unsigned i = 0; i < replicaCount; i++)
			{
				string replica = m_config->replicas[i];
				string login, password;

				FB_SIZE_T pos = replica.find('@');
				if (pos != string::npos)
				{
					const string temp = replica.substr(0, pos);
					replica = replica.substr(pos + 1);

					pos = temp.find(':');
					if (pos != string::npos)
					{
						login = temp.substr(0, pos);
						password = temp.substr(pos + 1);
					}
					else
					{
						login = temp;
					}
				}

				ClumpletWriter dpbBuffer(ClumpletReader::Tagged, MAX_SSHORT);

				if (org_dpb_length)
					dpbBuffer.reset(org_dpb, org_dpb_length);
				else
					dpbBuffer.reset(isc_dpb_version1);

				// remove remote protocol tags
				dpbBuffer.deleteWithTag(isc_dpb_address_path);
				dpbBuffer.deleteWithTag(isc_dpb_process_name);
				dpbBuffer.deleteWithTag(isc_dpb_process_id);
				// remove alias name
				dpbBuffer.deleteWithTag(isc_dpb_org_filename);
				// remove unnecessary options
				dpbBuffer.deleteWithTag(isc_dpb_dbkey_scope);
				dpbBuffer.deleteWithTag(isc_dpb_no_garbage_collect);
				dpbBuffer.deleteWithTag(isc_dpb_no_db_triggers);
				dpbBuffer.deleteWithTag(isc_dpb_map_attach);
				dpbBuffer.deleteWithTag(isc_dpb_config);
				// remove unwanted connect-time actions
				dpbBuffer.deleteWithTag(isc_dpb_sweep);
				dpbBuffer.deleteWithTag(isc_dpb_verify);
				dpbBuffer.deleteWithTag(isc_dpb_activate_shadow);
				dpbBuffer.deleteWithTag(isc_dpb_delete_shadow);
				dpbBuffer.deleteWithTag(isc_dpb_shutdown);
				dpbBuffer.deleteWithTag(isc_dpb_shutdown_delay);
				dpbBuffer.deleteWithTag(isc_dpb_online);
				dpbBuffer.deleteWithTag(isc_dpb_set_db_replica);

				if (login.hasData())
				{
					dpbBuffer.deleteWithTag(isc_dpb_user_name);
					dpbBuffer.deleteWithTag(isc_dpb_password);
					dpbBuffer.deleteWithTag(isc_dpb_password_enc);
					dpbBuffer.deleteWithTag(isc_dpb_trusted_auth);
					dpbBuffer.deleteWithTag(isc_dpb_auth_block);
					dpbBuffer.deleteWithTag(isc_dpb_sql_role_name);
					dpbBuffer.deleteWithTag(isc_dpb_trusted_role);

					dpbBuffer.insertString(isc_dpb_user_name, login);

					if (password.hasData())
						dpbBuffer.insertString(isc_dpb_password, password);
				}

				if (create_flag)
					dpbBuffer.insertString(isc_dpb_set_db_replica, guidStr);

				SSHORT dpbLength = (SSHORT) dpbBuffer.getBufferLength();
				SCHAR* dpb = (SCHAR*) dpbBuffer.getBuffer();

				ISC_STATUS_ARRAY temp_status;
				FB_API_HANDLE handle = 0;

				if (create_flag)
				{
					isc_create_database(temp_status, 0, replica.c_str(),
										&handle, dpbLength, dpb, 0);
				}
				else
				{
					isc_attach_database(temp_status, 0, replica.c_str(),
										&handle, dpbLength, dpb);
				}

				if (temp_status[1])
				{
					if (m_config->disableOnError)
					{
						errorCount++;
						m_slaves.add(0);
						logError(temp_status);
						continue;
					}

					Arg::StatusVector org_error(temp_status);
					Arg::StatusVector new_error;
					new_error << Arg::Gds(isc_random) << Arg::Str("Replication error");
					new_error.append(org_error);
					new_error.copyTo(temp_status);

					status_exception::raise(temp_status);
				}

				m_slaves.add(handle);
				m_manager->init(handle);
			}

			const unsigned slaveCount = (unsigned) m_slaves.getCount();

			if (slaveCount == errorCount)
				raiseError("No replicas are available");
			else if (errorCount)
				logMessage("One or more replicas are unavailable");

			ULONG flags = 0;
			if (m_config->compressRecords)
				flags |= FLAG_COMPRESS_RECORDS;
			if (m_config->masterPriority)
				flags |= FLAG_MASTER_PRIORITY;

			fb_assert(m_writer.isEmpty());

			m_writer.putTag(eInitialize);
			m_writer.putInt(PROTOCOL_VERSION2);
			m_writer.putInt(flags);
			m_writer.putString(guidStr);

			fb_assert(m_writer.getLength() == INIT_PACKET_SIZE);
		}

		void flush(FlushReason reason, bool force = false)
		{
			const bool sync = (reason == FLUSH_SYNC);

			ULONG length = m_writer.getLength();
			const UCHAR* const ptr = m_writer.getData();

			if (length)
			{
				FbLocalStatus local_status;
				ISC_STATUS_ARRAY temp_status;

				unsigned slaves = 0, errors = 0;

				for (unsigned i = 0; i < m_slaves.getCount(); i++)
				{
					if (m_slaves[i])
					{
						try
						{
							slaves++;
							m_manager->flush(m_slaves[i], length, ptr, sync, force);
						}
						catch (const Exception& ex)
						{
							if (m_config->disableOnError)
							{
								ex.stuffException(&local_status);
								errors++;
								logError(local_status->getErrors());
								m_manager->cleanup(m_slaves[i]);
								isc_detach_database(temp_status, &m_slaves[i]);
								slaves--;
								continue;
							}

							m_writer.clear();
							throw;
						}
					}
				}

				if (m_config->isLoggingActive())
				{
					if (length)
					{
						try
						{
							slaves++;
							m_manager->flush(0, length, ptr, sync, force);
						}
						catch (const Exception&)
						{
							m_writer.clear();
							throw;
						}
					}
				}

				if (!slaves)
					raiseError("No replicas are available");
				else if (errors)
					logMessage("One or more replicas are unavailable");

				m_bytes += length;
				m_writer.clear();
			}
		}

		void logError(const ISC_STATUS* status)
		{
			string message;

			char temp[BUFFER_LARGE];
			const ISC_STATUS* status_ptr = status;
			while (fb_interpret(temp, sizeof(temp), &status_ptr))
			{
				if (!message.isEmpty())
					message += "\n\t";

				message += temp;
			}

			logMessage(message);
		}

		void postError(FbStatusVector* status, const Exception& ex)
		{
			FbLocalStatus temp_status;
			ex.stuffException(&temp_status);
			logError(temp_status->getErrors());

			if (!m_alerted && m_config->alertCommand.hasData())
			{
				system(m_config->alertCommand.c_str());
				m_alerted = true;
			}

			if (m_config->disableOnError)
			{
				logMessage(SYNC_WARNING);
			}
			else
			{
				Arg::StatusVector org_error(&temp_status);
				Arg::StatusVector new_error;
				new_error << Arg::Gds(isc_random) << Arg::Str("Replication error");
				new_error.append(org_error);
				new_error.copyTo(status);
			}
		}
	};

	class ReplicationSlave: private PermanentStorage, public ErrorLogger
	{
		typedef GenericMap<Pair<NonPooled<AttNumber, Attachment*> > > AttachmentMap;
		typedef GenericMap<Pair<NonPooled<TraNumber, jrd_tra*> > > TransactionMap;
		typedef HalfStaticArray<bid, 16> BlobList;

		class LocalThreadContext
		{
		public:
			LocalThreadContext(thread_db* tdbb, Attachment* att, jrd_tra* tra = NULL, jrd_req* req = NULL)
				: m_tdbb(tdbb), m_att(tdbb->getAttachment())
			{
				tdbb->setAttachment(att);
				tdbb->setTransaction(tra);
				tdbb->setRequest(req);
			}

			~LocalThreadContext()
			{
				m_tdbb->setAttachment(m_att);
				m_tdbb->setTransaction(NULL);
				m_tdbb->setRequest(NULL);
			}

		private:
			thread_db* m_tdbb;
			Attachment* const m_att;
		};

	public:
		explicit ReplicationSlave(MemoryPool& pool, const PathName& filename)
			: PermanentStorage(pool),
			  ErrorLogger(pool, filename, SLAVE),
			  m_attMap(pool), m_txnMap(pool),
			  m_flags(0), m_longIds(false)
		{}

		void initialize(thread_db* tdbb, ULONG version, ULONG flags, const char* db_id);

		void cleanupAttachment(thread_db* tdbb, AttNumber att_id);

		void startTransaction(thread_db* tdbb, AttNumber att_id, TraNumber txn_id);
		void prepareTransaction(thread_db* tdbb, TraNumber txn_id);
		void commitTransaction(thread_db* tdbb, TraNumber txn_id);
		void rollbackTransaction(thread_db* tdbb, TraNumber txn_id);

		void startSavepoint(thread_db* tdbb, TraNumber txn_id);
		void cleanupSavepoint(thread_db* tdbb, TraNumber txn_id, bool undo);

		void insertRecord(thread_db* tdbb, TraNumber txn_id, const char* rel_name,
							ULONG length, const UCHAR* data);
		void updateRecord(thread_db* tdbb, TraNumber txn_id, const char* rel_name,
							ULONG org_length, const UCHAR* org_data,
							ULONG new_length, const UCHAR* new_data);
		void deleteRecord(thread_db* tdbb, TraNumber txn_id, const char* rel_name,
							ULONG length, const UCHAR* data);

		void changeGenerator(thread_db* tdbb, SLONG gen_id, SINT64 value);

		void storeBlob(thread_db* tdbb, TraNumber txn_id, bid* blob_id,
						 ULONG length, const UCHAR* data);

		void suspendTxns(thread_db* tdbb, TraNumber txn_id);
		void resumeTxns(thread_db* tdbb);
		void syncTxns(thread_db* tdbb, ULONG timeout);

		void shutdown(thread_db* tdbb);

		void postError(FbStatusVector* status, const Exception& ex)
		{
			FbLocalStatus temp_status;
			ex.stuffException(&temp_status);

			string message;

			char temp[BUFFER_LARGE];
			const ISC_STATUS* temp_status_ptr = temp_status->getErrors();
			while (fb_interpret(temp, sizeof(temp), &temp_status_ptr))
			{
				if (!message.isEmpty())
					message += "\n\t";

				message += temp;
			}

			logMessage(message);

			Arg::StatusVector org_error(&temp_status);
			Arg::StatusVector new_error;
			new_error << Arg::Gds(isc_random) << Arg::Str("Replication error");
			new_error.append(org_error);
			new_error.copyTo(status);
		}

		bool hasLongIds() const
		{
			return m_longIds;
		}

	private:
		RWLock m_lock;
		AttachmentMap m_attMap;
		TransactionMap m_txnMap;
		ULONG m_flags;
		Guid m_guid;
		bool m_longIds;

		Attachment* attach(thread_db* tdbb, AttNumber att_id);

		bool lookupKey(thread_db* tdbb, jrd_rel* relation, index_desc& idx);
		bool compareKey(thread_db* tdbb, jrd_rel* relation,
										  const index_desc& idx,
										  Record* record1, Record* record2);
		bool lookupRecord(thread_db* tdbb, jrd_rel* relation,
						  Record* record, RecordBitmap* bitmap,
						  index_desc& idx);

		void doInsert(thread_db* tdbb, record_param* rpb,
						jrd_tra* transaction);
		void doUpdate(thread_db* tdbb, record_param* org_rpb, record_param* new_rpb,
						jrd_tra* transaction, BlobList* blobs);
		void doDelete(thread_db* tdbb, record_param* rpb,
						jrd_tra* transaction);

		void logWarning(const char* msg, ...)
		{
#ifdef LOG_WARNINGS
			char buffer[BUFFER_LARGE];

			va_list ptr;
			va_start(ptr, msg);
			vsprintf(buffer, msg, ptr);
			va_end(ptr);

			string message;
			message.printf("WARNING: %s", buffer);
			logMessage(message);
#endif
		}
	};

} // namespace

// Master implementation

bool ReplicationMaster::cleanupAttachment(FbStatusVector* status, AttNumber att_id)
{
	try
	{
		MutexLockGuard guard(m_mutex, FB_FUNCTION);

		FB_SIZE_T pos;
		if (m_sessions.find(att_id, pos))
		{
			m_writer.putTag(eCleanupAttachment);
			m_writer.putId(att_id);
			flush(FLUSH_SYNC, true);

			m_sessions.remove(pos);
		}
	}
	catch (const Exception& ex)
	{
		postError(status, ex);
		return false;
	}

	return true;
}

bool ReplicationMaster::startTransaction(FbStatusVector* status, AttNumber att_id, TraNumber txn_id)
{
	try
	{
		MutexLockGuard guard(m_mutex, FB_FUNCTION);

		if (txn_id == 0)
			raiseError("Cannot replicate actions of system transaction");

		if (m_map.exist(txn_id))
			raiseError("Transaction %d already exists", txn_id);

		MemoryPool& pool = getPool();
		TransactionData* const txnData = FB_NEW_POOL(pool) TransactionData;
		txnData->buffer = FB_NEW_POOL(pool) Writer(pool, m_config->bufferSize);
		txnData->stack = FB_NEW_POOL(pool) SavepointStack(pool);
		txnData->flushes = 0;
		txnData->session = att_id;

		m_map.put(txn_id, txnData);

		Writer* const txnBuffer = txnData->buffer;

		txnBuffer->putTag(eStartTransaction);
		txnBuffer->putId(att_id);
		txnBuffer->putId(txn_id);
		fb_assert(txnBuffer->getLength() == START_TXN_PACKET_SIZE);
	}
	catch (const Exception& ex)
	{
		postError(status, ex);
		return false;
	}

	return true;
}

bool ReplicationMaster::prepareTransaction(FbStatusVector* status, TraNumber txn_id)
{
	try
	{
		MutexLockGuard guard(m_mutex, FB_FUNCTION);

		if (txn_id == 0)
			raiseError("Cannot replicate actions of system transaction");

		TransactionData* txnData = NULL;
		if (!m_map.get(txn_id, txnData))
			raiseError("Transaction %d is not found", txn_id);

		Writer* const txnBuffer = txnData->buffer;

		if (!txnData->flushes && txnBuffer->getLength() == START_TXN_PACKET_SIZE)
			return true; // no-op

		txnBuffer->putTag(ePrepareTransaction);
		txnBuffer->putId(txn_id);

		txnBuffer->mergeTo(m_writer);
		txnData->flushes++;
		flush(FLUSH_PREPARE);

		if (!m_sessions.exist(txnData->session))
			m_sessions.add(txnData->session);
	}
	catch (const Exception& ex)
	{
		postError(status, ex);
		return false;
	}

	return true;
}

bool ReplicationMaster::commitTransaction(FbStatusVector* status, TraNumber txn_id)
{
	try
	{
		MutexLockGuard guard(m_mutex, FB_FUNCTION);

		if (txn_id == 0)
			raiseError("Cannot replicate actions of system transaction");

		TransactionData* txnData = NULL;
		if (!m_map.get(txn_id, txnData))
			raiseError("Transaction %d is not found", txn_id);

		Writer* const txnBuffer = txnData->buffer;

		if (!txnData->flushes && txnBuffer->getLength() == START_TXN_PACKET_SIZE)
		{
			m_map.remove(txn_id);
			delete txnData->buffer;
			delete txnData->stack;
			delete txnData;

			return true;
		}

		fb_assert(txnData->flushes);
		fb_assert(txnBuffer->isEmpty());

		txnBuffer->putTag(eCommitTransaction);
		txnBuffer->putId(txn_id);

		txnBuffer->mergeTo(m_writer);
		txnData->flushes++;
		flush(FLUSH_SYNC);

		if (!m_sessions.exist(txnData->session))
			m_sessions.add(txnData->session);

		m_map.remove(txn_id);
		delete txnData->buffer;
		delete txnData->stack;
		delete txnData;
	}
	catch (const Exception& ex)
	{
		postError(status, ex);
		return false;
	}

	return true;
}

bool ReplicationMaster::rollbackTransaction(FbStatusVector* status, TraNumber txn_id)
{
	try
	{
		MutexLockGuard guard(m_mutex, FB_FUNCTION);

		if (txn_id == 0)
			raiseError("Cannot replicate actions of system transaction");

		TransactionData* txnData = NULL;
		if (!m_map.get(txn_id, txnData))
			raiseError("Transaction %d is not found", txn_id);

		const bool doRollback = (txnData->flushes != 0);

		if (doRollback)
		{
			m_writer.putTag(eRollbackTransaction);
			m_writer.putId(txn_id);

			txnData->flushes++;
			flush(FLUSH_SYNC, true);

			if (!m_sessions.exist(txnData->session))
				m_sessions.add(txnData->session);
		}

		m_map.remove(txn_id);
		delete txnData->buffer;
		delete txnData->stack;
		delete txnData;
	}
	catch (const Exception& ex)
	{
		postError(status, ex);
		return false;
	}

	return true;
}

bool ReplicationMaster::cleanupTransactions(FbStatusVector* status)
{
	try
	{
		MutexLockGuard guard(m_mutex, FB_FUNCTION);

		m_writer.putTag(eCleanupTransactions);

		flush(FLUSH_SYNC);
	}
	catch (const Exception& ex)
	{
		postError(status, ex);
		return false;
	}

	return true;
}

 bool ReplicationMaster::startSavepoint(FbStatusVector* status,
									   TraNumber txn_id,
									   SavNumber sav_id)
{
	try
	{
		MutexLockGuard guard(m_mutex, FB_FUNCTION);

		if (txn_id == 0)
			raiseError("Cannot replicate actions of system transaction");

		TransactionData* txnData = NULL;
		if (!m_map.get(txn_id, txnData))
			raiseError("Transaction %d is not found", txn_id);

		if (txnData->stack->exist(sav_id))
			return true;

		Writer* const txnBuffer = txnData->buffer;

		txnBuffer->putTag(eStartSavepoint);
		txnBuffer->putId(txn_id);

		if (txnBuffer->getLength() > m_config->bufferSize)
		{
			txnBuffer->mergeTo(m_writer);
			txnData->flushes++;
			flush(FLUSH_OVERFLOW);

			if (!m_sessions.exist(txnData->session))
				m_sessions.add(txnData->session);
		}

		txnData->stack->add(sav_id);
	}
	catch (const Exception& ex)
	{
		postError(status, ex);
		return false;
	}

	return true;
}

bool ReplicationMaster::cleanupSavepoint(FbStatusVector* status,
										 TraNumber txn_id,
										 SavNumber sav_id,
										 bool undo)
{
	try
	{
		MutexLockGuard guard(m_mutex, FB_FUNCTION);

		if (txn_id == 0)
			raiseError("Cannot replicate actions of system transaction");

		TransactionData* txnData = NULL;
		if (!m_map.get(txn_id, txnData))
			raiseError("Transaction %d is not found", txn_id);

		FB_SIZE_T pos;
		if (!txnData->stack->find(sav_id, pos))
			return true;

		fb_assert(txnData->stack->back() == sav_id);

		Writer* const txnBuffer = txnData->buffer;

		txnBuffer->putTag(eCleanupSavepoint);
		txnBuffer->putId(txn_id);
		txnBuffer->putInt(undo ? 1 : 0);

		if (txnBuffer->getLength() > m_config->bufferSize || undo)
		{
			txnBuffer->mergeTo(m_writer);
			txnData->flushes++;
			flush(undo ? FLUSH_SYNC : FLUSH_OVERFLOW);

			if (!m_sessions.exist(txnData->session))
				m_sessions.add(txnData->session);
		}

		txnData->stack->remove(pos);
	}
	catch (const Exception& ex)
	{
		postError(status, ex);
		return false;
	}

	return true;
}

bool ReplicationMaster::insertRecord(FbStatusVector* status,
									 TraNumber txn_id,
									 const char* rel_name,
									 ULONG length, const UCHAR* data)
{
	try
	{
		MutexLockGuard guard(m_mutex, FB_FUNCTION);

		if (txn_id == 0)
			raiseError("Cannot replicate actions of system transaction");

		TransactionData* txnData = NULL;
		if (!m_map.get(txn_id, txnData))
			raiseError("Transaction %d is not found", txn_id);

		UCharBuffer buffer(getPool());
		if (m_config->compressRecords)
		{
			const Compressor dcc(getPool(), length, data);
			length = (ULONG) dcc.getPackedLength();
			dcc.pack(data, buffer.getBuffer(length));
			data = buffer.begin();
		}

		Writer* const txnBuffer = txnData->buffer;

		txnBuffer->putTag(eInsertRecord);
		txnBuffer->putId(txn_id);
		txnBuffer->putString(rel_name);
		txnBuffer->putBinary(length, data);

		if (txnBuffer->getLength() > m_config->bufferSize)
		{
			txnBuffer->mergeTo(m_writer);
			txnData->flushes++;
			flush(FLUSH_OVERFLOW);

			if (!m_sessions.exist(txnData->session))
				m_sessions.add(txnData->session);
		}
	}
	catch (const Exception& ex)
	{
		postError(status, ex);
		return false;
	}

	return true;
}

bool ReplicationMaster::updateRecord(FbStatusVector* status,
									 TraNumber txn_id,
									 const char* rel_name,
									 ULONG org_length, const UCHAR* org_data,
									 ULONG new_length, const UCHAR* new_data)
{
	try
	{
		MutexLockGuard guard(m_mutex, FB_FUNCTION);

		if (txn_id == 0)
			raiseError("Cannot replicate actions of system transaction");

		TransactionData* txnData = NULL;
		if (!m_map.get(txn_id, txnData))
			raiseError("Transaction %d is not found", txn_id);

		UCharBuffer buffer1(getPool()), buffer2(getPool());
		if (m_config->compressRecords)
		{
			const Compressor dcc1(getPool(), org_length, org_data);
			org_length = (ULONG) dcc1.getPackedLength();
			dcc1.pack(org_data, buffer1.getBuffer(org_length));
			org_data = buffer1.begin();

			const Compressor dcc2(getPool(), new_length, new_data);
			new_length = (ULONG) dcc2.getPackedLength();
			dcc2.pack(new_data, buffer2.getBuffer(new_length));
			new_data = buffer2.begin();
		}

		Writer* const txnBuffer = txnData->buffer;

		txnBuffer->putTag(eUpdateRecord);
		txnBuffer->putId(txn_id);
		txnBuffer->putString(rel_name);
		txnBuffer->putBinary(org_length, org_data);
		txnBuffer->putBinary(new_length, new_data);

		if (txnBuffer->getLength() > m_config->bufferSize)
		{
			txnBuffer->mergeTo(m_writer);
			txnData->flushes++;
			flush(FLUSH_OVERFLOW);

			if (!m_sessions.exist(txnData->session))
				m_sessions.add(txnData->session);
		}
	}
	catch (const Exception& ex)
	{
		postError(status, ex);
		return false;
	}

	return true;
}

bool ReplicationMaster::deleteRecord(FbStatusVector* status,
									 TraNumber txn_id,
									 const char* rel_name,
									 ULONG length, const UCHAR* data)
{
	try
	{
		MutexLockGuard guard(m_mutex, FB_FUNCTION);

		if (txn_id == 0)
			raiseError("Cannot replicate actions of system transaction");

		TransactionData* txnData = NULL;
		if (!m_map.get(txn_id, txnData))
			raiseError("Transaction %d is not found", txn_id);

		UCharBuffer buffer(getPool());
		if (m_config->compressRecords)
		{
			const Compressor dcc(getPool(), length, data);
			length = (ULONG) dcc.getPackedLength();
			dcc.pack(data, buffer.getBuffer(length));
			data = buffer.begin();
		}

		Writer* const txnBuffer = txnData->buffer;

		txnBuffer->putTag(eDeleteRecord);
		txnBuffer->putId(txn_id);
		txnBuffer->putString(rel_name);
		txnBuffer->putBinary(length, data);

		if (txnBuffer->getLength() > m_config->bufferSize)
		{
			txnBuffer->mergeTo(m_writer);
			txnData->flushes++;
			flush(FLUSH_OVERFLOW);

			if (!m_sessions.exist(txnData->session))
				m_sessions.add(txnData->session);
		}
	}
	catch (const Exception& ex)
	{
		postError(status, ex);
		return false;
	}

	return true;
}

bool ReplicationMaster::changeGenerator(FbStatusVector* status,
										SLONG gen_id,
										SINT64 value)
{
	try
	{
		MutexLockGuard guard(m_mutex, FB_FUNCTION);

		m_writer.putTag(eChangeGenerator);
		m_writer.putInt(gen_id);
		m_writer.putBigInt(value);

		if (m_writer.getLength() > m_config->bufferSize)
			flush(FLUSH_OVERFLOW);
	}
	catch (const Exception& ex)
	{
		postError(status, ex);
		return false;
	}

	return true;
}

bool ReplicationMaster::storeBlob(FbStatusVector* status,
								  TraNumber txn_id, ISC_QUAD blob_id,
								  ULONG length, const UCHAR* data)
{
	try
	{
		MutexLockGuard guard(m_mutex, FB_FUNCTION);

		if (txn_id == 0)
			raiseError("Cannot replicate actions of system transaction");

		TransactionData* txnData = NULL;
		if (!m_map.get(txn_id, txnData))
			raiseError("Transaction %d is not found", txn_id);

		Writer* const txnBuffer = txnData->buffer;

		txnBuffer->putTag(eStoreBlob);
		txnBuffer->putId(txn_id);
		txnBuffer->putInt(blob_id.gds_quad_high);
		txnBuffer->putInt(blob_id.gds_quad_low);
		txnBuffer->putBinary(length, data);

		if (txnBuffer->getLength() > m_config->bufferSize)
		{
			txnBuffer->mergeTo(m_writer);
			txnData->flushes++;
			flush(FLUSH_OVERFLOW);

			if (!m_sessions.exist(txnData->session))
				m_sessions.add(txnData->session);
		}
	}
	catch (const Exception& ex)
	{
		postError(status, ex);
		return false;
	}

	return true;
}

void ReplicationMaster::shutdown(FbStatusVector* status, bool drop)
{
	try
	{
		MutexLockGuard guard(m_mutex, FB_FUNCTION);

/*		If someone thinks it's worth to replicate uncommitted changes
 * 		before disconnecting from the slave, feel free to add a
 * 		flush(FLUSH_SYNC) call here.
 */
		ISC_STATUS_ARRAY temp_status;

		m_writer.clear();

		for (FB_SIZE_T i = 0; i < m_slaves.getCount(); i++)
		{
			if (m_slaves[i])
			{
				if (drop)
					isc_drop_database(temp_status, &m_slaves[i]);
				else
					isc_detach_database(temp_status, &m_slaves[i]);

				if (temp_status[1])
					Arg::StatusVector(temp_status).raise();
			}
		}
	}
	catch (const Exception& ex)
	{
		postError(status, ex);
	}

	delete this;
}

bool ReplicationMaster::isTableEnabled(Jrd::thread_db* tdbb, Jrd::jrd_rel* const relation)
{
	try
	{
		MutexLockGuard guard(m_mutex, FB_FUNCTION);

		const MetaName tableName(relation->rel_name);
		bool enabled = false;
		if (!m_tables.get(tableName, enabled))
		{
			enabled = true;

			if (m_includeMatcher)
			{
				m_includeMatcher->reset();
				m_includeMatcher->process((const UCHAR*) tableName.c_str(), tableName.length());
				enabled = m_includeMatcher->result();
			}

			if (enabled && m_excludeMatcher)
			{
				m_excludeMatcher->reset();
				m_excludeMatcher->process((const UCHAR*) tableName.c_str(), tableName.length());
				enabled = !m_excludeMatcher->result();
			}

			if (enabled && m_config->excludeWithoutPK)
			{
				enabled = false;

				RelationPages* relPages = relation->getPages(tdbb);
				ULONG page = relPages->rel_index_root;

				if (page == 0 && relation->rel_id != 0)
				{
					DPM_scan_pages(tdbb);
					page = relPages->rel_index_root;
				}

				if (page)
				{
					const Database* dbb = tdbb->getDatabase();
					CHECK_DBB(dbb);

					WIN window(relPages->rel_pg_space_id, page);
					index_root_page* root = (index_root_page*) CCH_FETCH(tdbb, &window, LCK_read, pag_root);

					for (USHORT id = 0; id < root->irt_count; ++id)
					{
						const index_root_page::irt_repeat* irt_desc;
						irt_desc = root->irt_rpt + id;

						if (irt_desc->irt_flags & irt_unique)
						{
							enabled = true;
							break;
						}
					}

					CCH_RELEASE(tdbb, &window);
				}
			}

			m_tables.put(tableName, enabled);
		}

		return enabled;
	}
	catch (const Exception&)
	{
		// If we failed matching the table name due to some internal error, then
		// let's allow the table to be replicated. This is not a critical failure.
		return true;
	}
}

void ReplicationMaster::clearTableEnabled(const char* rel_name)
{
	MutexLockGuard guard(m_mutex, FB_FUNCTION);

	const MetaName tableName(rel_name);
	if (m_tables.exist(tableName))
		m_tables.remove(tableName);
}

bool ReplicationMaster::isMasterPriority()
{
	return m_config->masterPriority;
}


// Slave implementation

void ReplicationSlave::initialize(thread_db* tdbb, ULONG version, ULONG flags,
								  const char* guid_str)
{
#ifdef REPL_LOG
	string output;
	output.printf("OUTPUT: initialize (protocol version = %" ULONGFORMAT ", flags = %" ULONGFORMAT ")",
				  version, flags);
	logMessage(output);
#endif
#ifdef REPL_NOP
	return;
#endif

	if (version != PROTOCOL_VERSION1 && version != PROTOCOL_VERSION2)
		raiseError("Invalid protocol version %d", version);

	Database* const dbb = tdbb->getDatabase();

	if (!dbb->isReplica())
		raiseError("Database is not in the replica mode");

	Guid guid1, guid2;
	StringToGuid(&guid1, guid_str);

	USHORT length = sizeof(Guid);
	if (!PAG_get_clump(tdbb, Ods::HDR_repl_guid, &length, (UCHAR*) &guid2))
		raiseError("Database is not in the replica mode");

	if (memcmp(&guid1, &guid2, sizeof(Guid)))
		raiseError("Database cannot be replicated from different sources");

	m_guid = guid1;
	m_flags = flags;
	m_longIds = (version == PROTOCOL_VERSION2);
}

void ReplicationSlave::cleanupAttachment(thread_db* tdbb, AttNumber att_id)
{
#ifdef REPL_LOG
	string output;
	output.printf("OUTPUT: cleanup attachment (att_id = %" SQUADFORMAT ")", att_id);
	logMessage(output);
#endif
#ifdef REPL_NOP
	return;
#endif

	WriteLockGuard protect(m_lock, FB_FUNCTION);

	Attachment* attachment = NULL;
	if (!m_attMap.get(att_id, attachment))
		return;

	LocalThreadContext context(tdbb, attachment);
	Attachment::SyncGuard guard(attachment, FB_FUNCTION);

	while (attachment->att_transactions)
	{
		TraceTransactionEnd trace(attachment->att_transactions, false, false);
		TRA_release_transaction(tdbb, attachment->att_transactions, &trace);
	}

	StableAttachmentPart* const sAtt = attachment->getStable();

	{ // scope
		Mutex* const asyncMutex = sAtt->getMutex(true, true);
		MutexLockGuard asyncGuard(*asyncMutex, FB_FUNCTION);
		JRD_release_attachment(tdbb, attachment);
	}

	sAtt->release();

	m_attMap.remove(att_id);
}

void ReplicationSlave::startTransaction(thread_db* tdbb, AttNumber att_id, TraNumber txn_id)
{
#ifdef REPL_LOG
	string output;
	output.printf("OUTPUT: start transaction (att_id = %" SQUADFORMAT ", txn_id = %" SQUADFORMAT ")",
				  att_id, txn_id);
	logMessage(output);
#endif
#ifdef REPL_NOP
	return;
#endif

	WriteLockGuard protect(m_lock, FB_FUNCTION);

	Attachment* attachment = NULL;
	if (!m_attMap.get(att_id, attachment))
		attachment = attach(tdbb, att_id);

	if (m_txnMap.exist(txn_id))
		raiseError("Transaction %d already exists", txn_id);

	LocalThreadContext context(tdbb, attachment);
	Attachment::SyncGuard guard(attachment, FB_FUNCTION);

	jrd_tra* const transaction =
		TRA_start(tdbb, TRA_read_committed | TRA_rec_version | TRA_no_auto_undo, 1);

	m_txnMap.put(txn_id, transaction);
}

void ReplicationSlave::prepareTransaction(thread_db* tdbb, TraNumber txn_id)
{
#ifdef REPL_LOG
	string output;
	output.printf("OUTPUT: prepare transaction (txn_id = %" SQUADFORMAT ")", txn_id);
	logMessage(output);
#endif
#ifdef REPL_NOP
	return;
#endif

	ReadLockGuard protect(m_lock, FB_FUNCTION);

	if (m_flags & FLAG_IGNORE_MISSING)
		return;

	jrd_tra* transaction = NULL;
	if (!m_txnMap.get(txn_id, transaction))
		raiseError("Transaction %d is not found", txn_id);

	Attachment* const attachment = transaction->tra_attachment;

	LocalThreadContext context(tdbb, attachment, transaction);
	Attachment::SyncGuard guard(attachment, FB_FUNCTION);

	TRA_prepare(tdbb, transaction, 0, NULL);
}

void ReplicationSlave::commitTransaction(thread_db* tdbb, TraNumber txn_id)
{
#ifdef REPL_LOG
	string output;
	output.printf("OUTPUT: commit transaction (txn_id = %" SQUADFORMAT ")", txn_id);
	logMessage(output);
#endif
#ifdef REPL_NOP
	return;
#endif

	WriteLockGuard protect(m_lock, FB_FUNCTION);

	jrd_tra* transaction = NULL;
	if (!m_txnMap.get(txn_id, transaction))
	{
		if (m_flags & FLAG_IGNORE_MISSING)
			return;

		raiseError("Transaction %d is not found", txn_id);
	}

	Attachment* const attachment = transaction->tra_attachment;

	LocalThreadContext context(tdbb, attachment, transaction);
	Attachment::SyncGuard guard(attachment, FB_FUNCTION);

	TRA_commit(tdbb, transaction, false);

	m_txnMap.remove(txn_id);
}

void ReplicationSlave::rollbackTransaction(thread_db* tdbb, TraNumber txn_id)
{
#ifdef REPL_LOG
	string output;
	output.printf("OUTPUT: rollback transaction (txn_id = %" SQUADFORMAT ")", txn_id);
	logMessage(output);
#endif
#ifdef REPL_NOP
	return;
#endif

	WriteLockGuard protect(m_lock, FB_FUNCTION);

	jrd_tra* transaction = NULL;
	if (!m_txnMap.get(txn_id, transaction))
	{
		if (m_flags & FLAG_IGNORE_MISSING)
			return;

		raiseError("Transaction %d is not found", txn_id);
	}

	Attachment* const attachment = transaction->tra_attachment;

	LocalThreadContext context(tdbb, attachment, transaction);
	Attachment::SyncGuard guard(attachment, FB_FUNCTION);

	TRA_rollback(tdbb, transaction, false, true);

	m_txnMap.remove(txn_id);
}

void ReplicationSlave::startSavepoint(thread_db* tdbb, TraNumber txn_id)
{
#ifdef REPL_LOG
	string output;
	output.printf("OUTPUT: start savepoint (txn_id = %" SQUADFORMAT ")", txn_id);
	logMessage(output);
#endif
#ifdef REPL_NOP
	return;
#endif

	ReadLockGuard protect(m_lock, FB_FUNCTION);

	jrd_tra* transaction = NULL;
	if (!m_txnMap.get(txn_id, transaction))
	{
		if (m_flags & FLAG_IGNORE_MISSING)
			return;

		raiseError("Transaction %d is not found", txn_id);
	}

	Attachment* const attachment = transaction->tra_attachment;

	LocalThreadContext context(tdbb, attachment, transaction);
	Attachment::SyncGuard guard(attachment, FB_FUNCTION);

	VIO_start_save_point(tdbb, transaction);
}

void ReplicationSlave::cleanupSavepoint(thread_db* tdbb, TraNumber txn_id, bool undo)
{
#ifdef REPL_LOG
	string output;
	output.printf("OUTPUT: cleanup savepoint (txn_id = %" SQUADFORMAT ", undo = %" ULONGFORMAT ")",
				  txn_id, undo);
	logMessage(output);
#endif
#ifdef REPL_NOP
	return;
#endif

	ReadLockGuard protect(m_lock, FB_FUNCTION);

	jrd_tra* transaction = NULL;
	if (!m_txnMap.get(txn_id, transaction))
	{
		if (m_flags & FLAG_IGNORE_MISSING)
			return;

		raiseError("Transaction %d is not found", txn_id);
	}

	Attachment* const attachment = transaction->tra_attachment;

	LocalThreadContext context(tdbb, attachment, transaction);
	Attachment::SyncGuard guard(attachment, FB_FUNCTION);

	if (!transaction->tra_save_point)
		raiseError("Transaction %d has no savepoints to cleanup", txn_id);

	transaction->tra_save_point->sav_verb_count = (undo ? 1 : 0);
	VIO_verb_cleanup(tdbb, transaction);
}

void ReplicationSlave::insertRecord(thread_db* tdbb, TraNumber txn_id, const char* rel_name,
										 ULONG length, const UCHAR* data)
{
#ifdef REPL_LOG
	string output;
	output.printf("OUTPUT: insert record (txn_id = %" SQUADFORMAT ", table = %s"
				  ", length = %" ULONGFORMAT ")",
				  txn_id, rel_name, length);
	logMessage(output);
#endif
#ifdef REPL_NOP
	return;
#endif

	ReadLockGuard protect(m_lock, FB_FUNCTION);

	jrd_tra* transaction = NULL;
	if (!m_txnMap.get(txn_id, transaction))
	{
		if (m_flags & FLAG_IGNORE_MISSING)
			return;

		raiseError("Transaction %d is not found", txn_id);
	}

	Attachment* const attachment = transaction->tra_attachment;
	jrd_req* const request = attachment->att_repl_request;

	LocalThreadContext context(tdbb, attachment, transaction, request);
	Attachment::SyncGuard guard(attachment, FB_FUNCTION);

	TRA_attach_request(transaction, request);

	jrd_rel* const relation = MET_lookup_relation(tdbb, rel_name);
	if (!relation)
		raiseError("Table %s is not found", rel_name);

	if (!(relation->rel_flags & REL_scanned))
		MET_scan_relation(tdbb, relation);

	Vector<UCHAR, MAX_RECORD_SIZE> buffer;
	if (m_flags & FLAG_COMPRESS_RECORDS)
	{
		UCHAR* const ptr = buffer.begin();
		const ULONG capacity = (ULONG) buffer.getCapacity();
		length = Compressor::unpack(length, data, capacity, ptr) - ptr;
		data = ptr;
	}

	const Format* const format = MET_current(tdbb, relation);
	if (format->fmt_length != length)
	{
		raiseError("Record length mismatch for table %s: expected %d, found %d",
				   rel_name, format->fmt_length, length);
	}

	const USHORT rel_id = relation->rel_id;

	record_param rpb;
	rpb.rpb_relation = relation;

	rpb.rpb_record = attachment->att_repl_record;
	Record* const record = attachment->att_repl_record =
		VIO_record(tdbb, &rpb, format, request->req_pool);

	rpb.rpb_format_number = format->fmt_version;
	rpb.rpb_address = record->getData();
	rpb.rpb_length = length;
	record->copyDataFrom(data);

	bool found = false;

	if (m_flags & FLAG_MASTER_PRIORITY)
	{
		RecordBitmap* const bitmap = attachment->att_repl_bitmap;

		index_desc idx;
		const bool indexed = lookupRecord(tdbb, relation, record, bitmap, idx);

		if (bitmap->getFirst())
		{
			record_param temp_rpb = rpb;

			do {
				temp_rpb.rpb_number.setValue(bitmap->current());
				temp_rpb.rpb_record = NULL;

				if (VIO_get(tdbb, &temp_rpb, transaction, request->req_pool) &&
					(!indexed || compareKey(tdbb, relation, idx, record, temp_rpb.rpb_record)))
				{
					if (found)
						raiseError("Record in table %s is ambiguously identified using the primary/unique key", rel_name);

					rpb = temp_rpb;
					found = true;
				}
			} while (bitmap->getNext());

			if (!found)
			{
				delete temp_rpb.rpb_record;

				// Re-init the record to recover after being possibly changed inside VIO_get()

				VIO_record(tdbb, &rpb, format, request->req_pool);

				rpb.rpb_format_number = format->fmt_version;
				rpb.rpb_address = record->getData();
				rpb.rpb_length = length;
				record->copyDataFrom(data);
			}
		}
	}

	if (found)
	{
		logWarning("Record being inserted into table %s already exists, attempting to update", rel_name);

		record_param new_rpb = rpb;

		new_rpb.rpb_record = NULL;
		AutoPtr<Record> new_record(VIO_record(tdbb, &new_rpb, format, request->req_pool));

		new_rpb.rpb_format_number = format->fmt_version;
		new_rpb.rpb_address = new_record->getData();
		new_rpb.rpb_length = length;
		new_record->copyDataFrom(data);

		doUpdate(tdbb, &rpb, &new_rpb, transaction, NULL);
	}
	else
	{
		doInsert(tdbb, &rpb, transaction);
	}
}

void ReplicationSlave::updateRecord(thread_db* tdbb, TraNumber txn_id, const char* rel_name,
									ULONG org_length, const UCHAR* org_data,
									ULONG new_length, const UCHAR* new_data)
{
#ifdef REPL_LOG
	string output;
	output.printf("OUTPUT: update record (txn_id = %" SQUADFORMAT ", table = %s"
				  ", org_length = %" ULONGFORMAT ", new_length = %" ULONGFORMAT ")",
				  txn_id, rel_name, org_length, new_length);
	logMessage(output);
#endif
#ifdef REPL_NOP
	return;
#endif

	ReadLockGuard protect(m_lock, FB_FUNCTION);

	jrd_tra* transaction = NULL;
	if (!m_txnMap.get(txn_id, transaction))
	{
		if (m_flags & FLAG_IGNORE_MISSING)
			return;

		raiseError("Transaction %d is not found", txn_id);
	}

	Attachment* const attachment = transaction->tra_attachment;
	jrd_req* const request = attachment->att_repl_request;

	LocalThreadContext context(tdbb, attachment, transaction, request);
	Attachment::SyncGuard guard(attachment, FB_FUNCTION);

	TRA_attach_request(transaction, request);

	jrd_rel* const relation = MET_lookup_relation(tdbb, rel_name);
	if (!relation)
		raiseError("Table %s is not found", rel_name);

	fb_assert(relation->rel_name == rel_name);

	if (!(relation->rel_flags & REL_scanned))
		MET_scan_relation(tdbb, relation);

	Vector<UCHAR, MAX_RECORD_SIZE> buffer1, buffer2;
	if (m_flags & FLAG_COMPRESS_RECORDS)
	{
		UCHAR* const ptr1 = buffer1.begin();
		const ULONG capacity1 = (ULONG) buffer1.getCapacity();
		org_length = Compressor::unpack(org_length, org_data, capacity1, ptr1) - ptr1;
		org_data = ptr1;

		UCHAR* const ptr2 = buffer2.begin();
		const ULONG capacity2 = (ULONG) buffer2.getCapacity();
		new_length = Compressor::unpack(new_length, new_data, capacity2, ptr2) - ptr2;
		new_data = ptr2;
	}

	const Format* const format = MET_current(tdbb, relation);
	if (format->fmt_length != org_length)
	{
		raiseError("Record length mismatch for table %s: expected %d, found %d",
				   rel_name, format->fmt_length, org_length);
	}
	if (format->fmt_length != new_length)
	{
		raiseError("Record length mismatch for table %s: expected %d, found %d",
				   rel_name, format->fmt_length, new_length);
	}

	const USHORT rel_id = relation->rel_id;

	record_param org_rpb;
	org_rpb.rpb_relation = relation;

	org_rpb.rpb_record = attachment->att_repl_record;
	Record* const org_record = attachment->att_repl_record =
		VIO_record(tdbb, &org_rpb, format, request->req_pool);

	org_rpb.rpb_format_number = format->fmt_version;
	org_rpb.rpb_address = org_record->getData();
	org_rpb.rpb_length = org_length;
	org_record->copyDataFrom(org_data);

	BlobList source_blobs(getPool());
	source_blobs.resize(format->fmt_count);
	for (USHORT id = 0; id < format->fmt_count; id++)
	{
		dsc desc;
		if (DTYPE_IS_BLOB(format->fmt_desc[id].dsc_dtype) &&
			EVL_field(NULL, org_record, id, &desc))
		{
			bid* const source = (bid*) desc.dsc_address;

			if (!source->isEmpty())
				source_blobs[id] = *source;
		}
	}

	RecordBitmap* const bitmap = attachment->att_repl_bitmap;

	index_desc idx;
	const bool indexed = lookupRecord(tdbb, relation, org_record, bitmap, idx);

	bool found = false;

	if (bitmap->getFirst())
	{
		record_param temp_rpb = org_rpb;

		do {
			temp_rpb.rpb_number.setValue(bitmap->current());
			temp_rpb.rpb_record = NULL;

			if (VIO_get(tdbb, &temp_rpb, transaction, request->req_pool) &&
				(!indexed || compareKey(tdbb, relation, idx, org_record, temp_rpb.rpb_record)))
			{
				if (found)
					raiseError("Record in table %s is ambiguously identified using the primary/unique key", rel_name);

				org_rpb = temp_rpb;
				found = true;
			}
		} while (bitmap->getNext());

		if (!found)
			delete temp_rpb.rpb_record;
	}

	record_param new_rpb;
	new_rpb.rpb_relation = relation;

	new_rpb.rpb_record = NULL;
	AutoPtr<Record> new_record(VIO_record(tdbb, &new_rpb, format, request->req_pool));

	new_rpb.rpb_format_number = format->fmt_version;
	new_rpb.rpb_address = new_record->getData();
	new_rpb.rpb_length = new_length;
	new_record->copyDataFrom(new_data);

	if (found)
	{
		index_desc idx;
		if (lookupKey(tdbb, relation, idx))
		{
			temporary_key org_key, new_key;

			idx_e result = BTR_key(tdbb, relation, org_record, &idx, &org_key, NULL, false);
			if (result != idx_e_ok)
			{
				IndexErrorContext context(relation, &idx);
				context.raise(tdbb, result, org_record);
			}

			result = BTR_key(tdbb, relation, new_record, &idx, &new_key, NULL, false);
			if (result != idx_e_ok)
			{
				IndexErrorContext context(relation, &idx);
				context.raise(tdbb, result, new_record);
			}

			if (org_key.key_length != new_key.key_length ||
				memcmp(org_key.key_data, new_key.key_data, org_key.key_length))
			{
				logWarning("Primary/unique key is changed during update for table %s", rel_name);
			}
		}

		doUpdate(tdbb, &org_rpb, &new_rpb, transaction, &source_blobs);
	}
	else if (m_flags & FLAG_MASTER_PRIORITY)
	{
		logWarning("Record being updated in table %s does not exist, attempting to insert", rel_name);
		doInsert(tdbb, &new_rpb, transaction);
	}
	else
	{
		raiseError("Record in table %s cannot be located via the primary/unique key", rel_name);
	}
}

void ReplicationSlave::deleteRecord(thread_db* tdbb, TraNumber txn_id, const char* rel_name,
									ULONG length, const UCHAR* data)
{
#ifdef REPL_LOG
	string output;
	output.printf("OUTPUT: delete record (txn_id = %" SQUADFORMAT ", table = %s"
				  ", length = %" ULONGFORMAT ")",
				  txn_id, rel_name, length);
	logMessage(output);
#endif
#ifdef REPL_NOP
	return;
#endif

	ReadLockGuard protect(m_lock, FB_FUNCTION);

	jrd_tra* transaction = NULL;
	if (!m_txnMap.get(txn_id, transaction))
	{
		if (m_flags & FLAG_IGNORE_MISSING)
			return;

		raiseError("Transaction %d is not found", txn_id);
	}

	Attachment* const attachment = transaction->tra_attachment;
	jrd_req* const request = attachment->att_repl_request;

	LocalThreadContext context(tdbb, attachment, transaction, request);
	Attachment::SyncGuard guard(attachment, FB_FUNCTION);

	TRA_attach_request(transaction, request);

	jrd_rel* const relation = MET_lookup_relation(tdbb, rel_name);
	if (!relation)
		raiseError("Table %s is not found", rel_name);

	if (!(relation->rel_flags & REL_scanned))
		MET_scan_relation(tdbb, relation);

	Vector<UCHAR, MAX_RECORD_SIZE> buffer;
	if (m_flags & FLAG_COMPRESS_RECORDS)
	{
		UCHAR* const ptr = buffer.begin();
		const ULONG capacity = (ULONG) buffer.getCapacity();
		length = Compressor::unpack(length, data, capacity, ptr) - ptr;
		data = ptr;
	}

	const Format* const format = MET_current(tdbb, relation);
	if (format->fmt_length != length)
	{
		raiseError("Record length mismatch for table %s: expected %d, found %d",
				   rel_name, format->fmt_length, length);
	}

	const USHORT rel_id = relation->rel_id;

	record_param rpb;
	rpb.rpb_relation = relation;

	rpb.rpb_record = attachment->att_repl_record;
	Record* const record = attachment->att_repl_record =
		VIO_record(tdbb, &rpb, format, request->req_pool);

	rpb.rpb_format_number = format->fmt_version;
	rpb.rpb_address = record->getData();
	rpb.rpb_length = length;
	record->copyDataFrom(data);

	RecordBitmap* const bitmap = attachment->att_repl_bitmap;

	index_desc idx;
	const bool indexed = lookupRecord(tdbb, relation, record, bitmap, idx);

	bool found = false;

	if (bitmap->getFirst())
	{
		record_param temp_rpb = rpb;

		do {
			temp_rpb.rpb_number.setValue(bitmap->current());
			temp_rpb.rpb_record = NULL;

			if (VIO_get(tdbb, &temp_rpb, transaction, request->req_pool) &&
				(!indexed || compareKey(tdbb, relation, idx, record, temp_rpb.rpb_record)))
			{
				if (found)
					raiseError("Record in table %s is ambiguously identified using the primary/unique key", rel_name);

				rpb = temp_rpb;
				found = true;
			}
		} while (bitmap->getNext());

		if (!found)
			delete temp_rpb.rpb_record;
	}

	if (found)
	{
		doDelete(tdbb, &rpb, transaction);
	}
	else if (m_flags & FLAG_MASTER_PRIORITY)
	{
		logWarning("Record being deleted from table %s does not exist, ignoring", rel_name);
	}
	else
	{
		raiseError("Record in table %s cannot be located via the primary/unique key", rel_name);
	}
}

void ReplicationSlave::changeGenerator(thread_db* tdbb, SLONG gen_id, SINT64 value)
{
#ifdef REPL_LOG
	string output;
	output.printf("OUTPUT: change generator (gen_id = %" SLONGFORMAT ", value = %" SQUADFORMAT ")",
				  gen_id, value);
	logMessage(output);
#endif
#ifdef REPL_NOP
	return;
#endif

	ReadLockGuard protect(m_lock, FB_FUNCTION);

	if (DPM_gen_id(tdbb, gen_id, false, 0) < value)
		DPM_gen_id(tdbb, gen_id, true, value);
}

void ReplicationSlave::storeBlob(thread_db* tdbb, TraNumber txn_id, bid* blob_id,
								 ULONG length, const UCHAR* data)
{
#ifdef REPL_LOG
	string output;
	output.printf("OUTPUT: store blob (txn_id = %" SQUADFORMAT ", rel_id = %" SLONGFORMAT
				  ", length = %" ULONGFORMAT ")",
				  txn_id, blob_id->bid_internal.bid_relation_id, length);
	logMessage(output);
#endif
#ifdef REPL_NOP
	return;
#endif

	ReadLockGuard protect(m_lock, FB_FUNCTION);

	jrd_tra* transaction = NULL;
	if (!m_txnMap.get(txn_id, transaction))
	{
		if (m_flags & FLAG_IGNORE_MISSING)
			return;

		raiseError("Transaction %d is not found", txn_id);
	}

	Attachment* const attachment = transaction->tra_attachment;

	LocalThreadContext context(tdbb, attachment, transaction);
	Attachment::SyncGuard guard(attachment, FB_FUNCTION);

	const bid org_blob_id = *blob_id;
	blb* const blob = blb::create(tdbb, transaction, blob_id);
	blob->blb_repl_id = org_blob_id;
	blob->BLB_put_data(tdbb, data, length);
	blob->BLB_close(tdbb);
}

void ReplicationSlave::shutdown(thread_db* tdbb)
{
	WriteLockGuard protect(m_lock, FB_FUNCTION);

	TransactionMap::Accessor txnAccessor(&m_txnMap);
	if (txnAccessor.getFirst())
	{
		do {
			jrd_tra* const transaction = txnAccessor.current()->second;
			Attachment* const attachment = transaction->tra_attachment;
			LocalThreadContext context(tdbb, attachment, transaction);
			Attachment::SyncGuard guard(attachment, FB_FUNCTION);
			TRA_rollback(tdbb, transaction, false, true);
		} while (txnAccessor.getNext());
	}

	AttachmentMap::Accessor attAccessor(&m_attMap);
	if (attAccessor.getFirst())
	{
		do {
			Attachment* const attachment = attAccessor.current()->second;
			LocalThreadContext context(tdbb, attachment);
			Attachment::SyncGuard guard(attachment, FB_FUNCTION);

			StableAttachmentPart* const sAtt = attachment->getStable();

			{ // scope
				Mutex* const asyncMutex = sAtt->getMutex(true, true);
				MutexLockGuard asyncGuard(*asyncMutex, FB_FUNCTION);
				JRD_release_attachment(tdbb, attachment);
			}

			sAtt->release();
		} while (attAccessor.getNext());
	}

	m_txnMap.clear();
	m_attMap.clear();
}

Attachment* ReplicationSlave::attach(thread_db* tdbb, AttNumber att_id)
{
	Database* const dbb = tdbb->getDatabase();

	Attachment* const baseAttachment = tdbb->getAttachment();
	Attachment* const attachment = Attachment::create(dbb);

	StableAttachmentPart* sAtt = FB_NEW StableAttachmentPart(attachment);
	attachment->setStable(sAtt);
	sAtt->addRef();

	LocalThreadContext context(tdbb, attachment);
	Attachment::SyncGuard guard(attachment, FB_FUNCTION);

	m_attMap.put(att_id, attachment);

	jrd_req* request = NULL;

	try
	{
		MemoryPool& att_pool = *attachment->att_pool;
		attachment->att_filename = baseAttachment->att_filename;
		attachment->att_user = FB_NEW_POOL(att_pool) UserId(att_pool, *baseAttachment->att_user);
		attachment->att_charset = baseAttachment->att_charset;
		attachment->att_flags = baseAttachment->att_flags;

		LCK_init(tdbb, LCK_OWNER_attachment);
		INI_init(tdbb);
		INI_init2(tdbb);
		PAG_header(tdbb, true);
		TRA_init(attachment);

		MemoryPool* const req_pool = attachment->createPool();
		Jrd::ContextPoolHolder context(tdbb, req_pool);
		AutoPtr<CompilerScratch> csb(CompilerScratch::newCsb(*req_pool, 0));

		request = JrdStatement::makeRequest(tdbb, csb, true);
		request->req_attachment = attachment;
		request->req_timestamp.validate();

		attachment->att_repl_request = request;
		attachment->att_repl_bitmap = FB_NEW_POOL(att_pool) RecordBitmap(att_pool);
	}
	catch (const Exception&)
	{
		if (request)
			CMP_release(tdbb, request);

		{ // scope
			Mutex* const asyncMutex = sAtt->getMutex(true, true);
			MutexLockGuard asyncGuard(*asyncMutex, FB_FUNCTION);
			JRD_release_attachment(tdbb, attachment);
		}

		sAtt->release();
		m_attMap.remove(att_id);

		throw;
	}

	return attachment;
}

bool ReplicationSlave::lookupKey(thread_db* tdbb, jrd_rel* relation, index_desc& key)
{
	RelationPages* const relPages = relation->getPages(tdbb);
	SLONG page = relPages->rel_index_root;
	if (!page)
	{
		DPM_scan_pages(tdbb);
		page = relPages->rel_index_root;
	}

	const PageNumber root_page(relPages->rel_pg_space_id, page);
	win window(root_page);
	index_root_page* const root = (index_root_page*) CCH_FETCH(tdbb, &window, LCK_read, pag_root);

	index_desc idx;
	idx.idx_id = key.idx_id = idx_invalid;

	for (USHORT i = 0; i < root->irt_count; i++)
	{
		if (BTR_description(tdbb, relation, root, &idx, i))
		{
			if (idx.idx_flags & idx_primary)
			{
				key = idx;
				break;
			}

			if (idx.idx_flags & idx_unique)
			{
				if ((key.idx_id == idx_invalid) || (idx.idx_count < key.idx_count))
				{
					key = idx;
				}
			}
		}
	}

	CCH_RELEASE(tdbb, &window);

	return (key.idx_id != idx_invalid);
}

bool ReplicationSlave::compareKey(thread_db* tdbb, jrd_rel* relation, const index_desc& idx,
								  Record* record1, Record* record2)
{
	bool equal = true;

	for (USHORT i = 0; i < idx.idx_count; i++)
	{
		const USHORT field_id = idx.idx_rpt[i].idx_field;

		dsc desc1, desc2;

		const bool null1 = !EVL_field(relation, record1, field_id, &desc1);
		const bool null2 = !EVL_field(relation, record2, field_id, &desc2);

		if (null1 != null2 || (!null1 && MOV_compare(&desc1, &desc2)))
		{
			equal = false;
			break;
		}
	}

	return equal;
}

bool ReplicationSlave::lookupRecord(thread_db* tdbb,
									jrd_rel* relation, Record* record,
									RecordBitmap* bitmap,
									index_desc& idx)
{
	RecordBitmap::reset(bitmap);

	// Special case: RDB$DATABASE has no keys but it's guaranteed to have only one record
	if (relation->rel_id == rel_database)
	{
		bitmap->set(0);
		return false;
	}

	if (lookupKey(tdbb, relation, idx))
	{
		temporary_key key;
		const idx_e result = BTR_key(tdbb, relation, record, &idx, &key, NULL, false);
		if (result != idx_e_ok)
		{
			IndexErrorContext context(relation, &idx);
			context.raise(tdbb, result, record);
		}

		IndexRetrieval retrieval(relation, &idx, idx.idx_count, &key);
		retrieval.irb_generic = irb_equality | (idx.idx_flags & idx_descending ? irb_descending : 0);

		BTR_evaluate(tdbb, &retrieval, &bitmap, NULL);
		return true;
	}

	NoKeyTable* table = NULL;

	for (unsigned i = 0; i < FB_NELEM(NO_KEY_TABLES); i++)
	{
		NoKeyTable* const tab = &NO_KEY_TABLES[i];

		if (tab->rel_id == relation->rel_id)
		{
			table = tab;
			break;
		}
	}

	if (!table)
		raiseError("Table %s has no unique key", relation->rel_name.c_str());

	jrd_tra* const transaction = tdbb->getTransaction();
	jrd_req* const request = tdbb->getRequest();

	RLCK_reserve_relation(tdbb, transaction, relation, false);

	record_param rpb;
	rpb.rpb_relation = relation;
	rpb.rpb_number.setValue(BOF_NUMBER);

	while (VIO_next_record(tdbb, &rpb, transaction, request->req_pool, false))
	{
		Record* const seq_record = rpb.rpb_record;
		fb_assert(seq_record);

		bool matched = true;

		for (unsigned i = 0; i < FB_NELEM(table->rel_fields); i++)
		{
			const USHORT field_id = table->rel_fields[i];

			if (field_id == MAX_USHORT)
				break;

			dsc desc1, desc2;

			const bool null1 = !EVL_field(relation, record, field_id, &desc1);
			const bool null2 = !EVL_field(relation, seq_record, field_id, &desc2);

			if (null1 != null2 || !null1 && MOV_compare(&desc1, &desc2))
			{
				matched = false;
				break;
			}
		}

		if (matched)
			bitmap->set(rpb.rpb_number.getValue());
	}

	delete rpb.rpb_record;
	return false;
}

void ReplicationSlave::doInsert(thread_db* tdbb, record_param* rpb,
								jrd_tra* transaction)
{
	Database* const dbb = tdbb->getDatabase();
	fb_assert(!(transaction->tra_flags & TRA_system));

	Record* const record = rpb->rpb_record;
	const Format* const format = record->getFormat();
	jrd_rel* const relation = rpb->rpb_relation;

	const char* const rel_name = relation->rel_name.c_str();

	RLCK_reserve_relation(tdbb, transaction, relation, true);

	for (USHORT id = 0; id < format->fmt_count; id++)
	{
		dsc desc;
		if (DTYPE_IS_BLOB(format->fmt_desc[id].dsc_dtype) &&
			EVL_field(NULL, record, id, &desc))
		{
			bid* const destination = (bid*) desc.dsc_address;

			if (!destination->isEmpty())
			{
				bool found = false;

				if (transaction->tra_blobs->getFirst())
				{
					do
					{
						BlobIndex* const current = &transaction->tra_blobs->current();

						if (!current->bli_materialized && current->bli_blob_object &&
							current->bli_blob_object->blb_repl_id == *destination)
						{
							blb* const blob = current->bli_blob_object;
							blob->blb_relation = relation;
							blob->blb_sub_type = desc.getBlobSubType();
							blob->blb_charset = desc.getCharSet();
							destination->set_permanent(relation->rel_id, DPM_store_blob(tdbb, blob, record));
							current->bli_materialized = true;
							current->bli_blob_id = *destination;
							transaction->tra_blobs->fastRemove();
							found = true;
							break;
						}
					}
					while (transaction->tra_blobs->getNext());
				}

				if (!found)
				{
					const ULONG num1 = destination->bid_quad.bid_quad_high;
					const ULONG num2 = destination->bid_quad.bid_quad_low;
					raiseError("Blob %d.%d is not found for table %s", num1, num2, rel_name);
				}
			}
		}
	}

	++transaction->tra_save_point->sav_verb_count;

	VIO_store(tdbb, rpb, transaction);
	IDX_store(tdbb, rpb, transaction);

	--transaction->tra_save_point->sav_verb_count;
}

void ReplicationSlave::doUpdate(thread_db* tdbb, record_param* org_rpb, record_param* new_rpb,
								jrd_tra* transaction, BlobList* blobs)
{
	Database* const dbb = tdbb->getDatabase();
	fb_assert(!(transaction->tra_flags & TRA_system));

	Record* const org_record = org_rpb->rpb_record;
	Record* const new_record = new_rpb->rpb_record;
	const Format* const format = new_record->getFormat();
	jrd_rel* const relation = new_rpb->rpb_relation;

	const char* const rel_name = relation->rel_name.c_str();

	RLCK_reserve_relation(tdbb, transaction, relation, true);

	for (USHORT id = 0; id < format->fmt_count; id++)
	{
		dsc desc;
		if (DTYPE_IS_BLOB(format->fmt_desc[id].dsc_dtype) &&
			EVL_field(NULL, new_record, id, &desc))
		{
			bid* const destination = (bid*) desc.dsc_address;
			bid* const source = (blobs && id < blobs->getCount()) ? (bid*) &(*blobs)[id] : NULL;

			if (!destination->isEmpty())
			{
				const bool same_blobs = (source && *destination == *source);

				if (same_blobs)
				{
					if (EVL_field(NULL, org_record, id, &desc))
						*destination = *(bid*) desc.dsc_address;
					else
						destination->clear();
				}
				else
				{
					bool found = false;

					if (transaction->tra_blobs->getFirst())
					{
						do
						{
							BlobIndex* const current = &transaction->tra_blobs->current();

							if (!current->bli_materialized && current->bli_blob_object &&
								current->bli_blob_object->blb_repl_id == *destination)
							{
								blb* const blob = current->bli_blob_object;
								blob->blb_relation = relation;
								blob->blb_sub_type = desc.getBlobSubType();
								blob->blb_charset = desc.getCharSet();
								destination->set_permanent(relation->rel_id, DPM_store_blob(tdbb, blob, new_record));
								current->bli_materialized = true;
								current->bli_blob_id = *destination;
								transaction->tra_blobs->fastRemove();
								found = true;
								break;
							}
						}
						while (transaction->tra_blobs->getNext());
					}

					if (!found)
					{
						const ULONG num1 = destination->bid_quad.bid_quad_high;
						const ULONG num2 = destination->bid_quad.bid_quad_low;
						raiseError("Blob %d.%d is not found for table %s", num1, num2, rel_name);
					}
				}
			}
		}
	}

	++transaction->tra_save_point->sav_verb_count;

	VIO_modify(tdbb, org_rpb, new_rpb, transaction);
	IDX_modify(tdbb, org_rpb, new_rpb, transaction);

	--transaction->tra_save_point->sav_verb_count;
}

void ReplicationSlave::doDelete(thread_db* tdbb, record_param* rpb,
								jrd_tra* transaction)
{
	Database* const dbb = tdbb->getDatabase();
	fb_assert(!(transaction->tra_flags & TRA_system));

	RLCK_reserve_relation(tdbb, transaction, rpb->rpb_relation, true);

	++transaction->tra_save_point->sav_verb_count;

	VIO_erase(tdbb, rpb, transaction);

	--transaction->tra_save_point->sav_verb_count;
}

namespace Jrd
{

void REPL_replicate(thread_db* tdbb, ULONG length, const UCHAR* data)
{
	Database* const dbb = tdbb->getDatabase();

	if (dbb->readOnly())
		raiseError("Replication is impossible for read-only database");

	Attachment* const attachment = tdbb->getAttachment();
	ReplicationSlave* slave = attachment->att_repl_slave;

	if (!slave)
	{
		MemoryPool& pool = *attachment->att_pool;
		slave = attachment->att_repl_slave =
			FB_NEW_POOL(pool) ReplicationSlave(pool, dbb->dbb_filename);
	}

#ifdef REPL_LOG
	string output;
	output.printf("OUTPUT: replicate packet (length = %" ULONGFORMAT ")", length);
	slave->logMessage(output);
#endif

	try {

	if (length && data[0] == eInitialize)
	{
		fb_assert(length >= INIT_PACKET_SIZE);

		Reader reader(INIT_PACKET_SIZE, data, false);

		reader.getTag(); // skip eInitialize
		const ULONG version = reader.getInt();
		const ULONG flags = reader.getInt();
		const string dbId = reader.getString();
		fb_assert(dbId.length() == GUID_STR_LENGTH);
		slave->initialize(tdbb, version, flags, dbId.c_str());

		length -= INIT_PACKET_SIZE;
		data += INIT_PACKET_SIZE;
	}

	Reader reader(length, data, slave->hasLongIds());

	while (reader.hasData())
	{
		const UCHAR op = reader.getTag();

		switch (op)
		{
		case eInitialize:
			{
				fb_assert(false);
				const ULONG version = reader.getInt();
				const ULONG flags = reader.getInt();
				const string dbId = reader.getString();
				fb_assert(dbId.length() == GUID_STR_LENGTH);
				slave->initialize(tdbb, version, flags, dbId.c_str());
			}
			break;

		case eCleanupAttachment:
			{
				const AttNumber att_id = reader.getId();
				slave->cleanupAttachment(tdbb, att_id);
			}
			break;

		case eStartTransaction:
			{
				const AttNumber att_id = reader.getId();
				const TraNumber tra_id = reader.getId();
				slave->startTransaction(tdbb, att_id, tra_id);
			}
			break;

		case ePrepareTransaction:
			{
				const TraNumber tra_id = reader.getId();
				slave->prepareTransaction(tdbb, tra_id);
			}
			break;

		case eCommitTransaction:
			{
				const TraNumber tra_id = reader.getId();
				slave->commitTransaction(tdbb, tra_id);
			}
			break;

		case eRollbackTransaction:
			{
				const TraNumber tra_id = reader.getId();
				slave->rollbackTransaction(tdbb, tra_id);
			}
			break;

		case eStartSavepoint:
			{
				const TraNumber tra_id = reader.getId();
				slave->startSavepoint(tdbb, tra_id);
			}
			break;

		case eCleanupSavepoint:
			{
				const TraNumber tra_id = reader.getId();
				const bool undo = reader.getInt() != 0;
				slave->cleanupSavepoint(tdbb, tra_id, undo);
			}
			break;

		case eInsertRecord:
			{
				const TraNumber tra_id = reader.getId();
				const string relName = reader.getString();
				const UCHAR* record = NULL;
				const ULONG length = reader.getBinary(record);
				slave->insertRecord(tdbb, tra_id, relName.c_str(), length, record);
			}
			break;

		case eUpdateRecord:
			{
				const TraNumber tra_id = reader.getId();
				const string relName = reader.getString();
				const UCHAR* org_record = NULL;
				const ULONG org_length = reader.getBinary(org_record);
				const UCHAR* new_record = NULL;
				const ULONG new_length = reader.getBinary(new_record);
				slave->updateRecord(tdbb, tra_id, relName.c_str(),
									org_length, org_record,
									new_length, new_record);
			}
			break;

		case eDeleteRecord:
			{
				const TraNumber tra_id = reader.getId();
				const string relName = reader.getString();
				const UCHAR* record = NULL;
				const ULONG length = reader.getBinary(record);
				slave->deleteRecord(tdbb, tra_id, relName.c_str(), length, record);
			}
			break;

		case eChangeGenerator:
			{
				const SLONG gen_id = reader.getInt();
				const SINT64 value = reader.getBigInt();
				slave->changeGenerator(tdbb, gen_id, value);
			}
			break;

		case eStoreBlob:
			{
				bid blob_id;
				const TraNumber tra_id = reader.getId();
				blob_id.bid_quad.bid_quad_high = reader.getInt();
				blob_id.bid_quad.bid_quad_low = reader.getInt();
				const UCHAR* blob = NULL;
				const ULONG length = reader.getBinary(blob);
				slave->storeBlob(tdbb, tra_id, &blob_id, length, blob);
			}
			break;

		case eCleanupTransactions:
			break;

		default:
			fb_assert(false);
		}
	}

	} // try
	catch (const Exception& ex)
	{
		slave->postError(tdbb->tdbb_status_vector, ex);
		throw;
	}
}

void REPL_init(thread_db* tdbb, IReplicationHelper* helper,
			   ULONG dpb_length, const UCHAR* dpb, bool create_flag)
{
	Attachment* const attachment = tdbb->getAttachment();
	MemoryPool& pool = *attachment->att_pool;

	try
	{
		attachment->att_replicator = (IReplicationMaster*)
			FB_NEW_POOL(pool) ReplicationMaster(pool, helper, dpb_length, dpb, create_flag);
	}
	catch (const LongJump&)
	{} // no-op
}

void REPL_shutdown(thread_db* tdbb)
{
	Attachment* const attachment = tdbb->getAttachment();

	if (attachment->att_repl_slave)
	{
		try
		{
			attachment->att_repl_slave->shutdown(tdbb);
		}
		catch (const Exception& ex)
		{
			attachment->att_repl_slave->postError(tdbb->tdbb_status_vector, ex);
			throw;
		}

		delete attachment->att_repl_slave;
		attachment->att_repl_slave = NULL;
	}
}

} // namespace
