/*
 *  The contents of this file are subject to the Initial
 *  Developer's Public License Version 1.0 (the "License");
 *  you may not use this file except in compliance with the
 *  License. You may obtain a copy of the License at
 *  http://www.ibphoenix.com/main.nfs?a=ibphoenix&page=ibp_idpl.
 *
 *  Software distributed under the License is distributed AS IS,
 *  WITHOUT WARRANTY OF ANY KIND, either express or implied.
 *  See the License for the specific language governing rights
 *  and limitations under the License.
 *
 *  The Original Code was created by Dmitry Yemanov
 *  for the Firebird Open Source RDBMS project.
 *
 *  Copyright (c) 2014 Dmitry Yemanov <dimitr@firebirdsql.org>
 *  and all contributors signed below.
 *
 *  All Rights Reserved.
 *  Contributor(s): ______________________________________.
 */

#include "firebird.h"
#include "../common/classes/alloc.h"
#include "../common/classes/auto.h"
#include "../common/isc_proto.h"
#include "../common/isc_s_proto.h"
#include "../common/os/os_utils.h"
#include "../common/os/path_utils.h"
#include "../jrd/jrd.h"

#include "ReplicationLog.h"
#include "ReplicationUtils.h"

#include <sys/types.h>
#include <sys/stat.h>

#ifdef HAVE_MMAP
#include <fcntl.h>
#include <sys/mman.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef WIN_NT
#include <process.h>
#include <io.h>
#include <fcntl.h>
#endif

#if defined(O_DSYNC)
#define SYNC		O_DSYNC
#elif defined(O_SYNC)
#define SYNC		O_SYNC
#elif defined(O_FSYNC)
#define SYNC		O_FSYNC
#else
#define SYNC		0
#endif

#ifndef O_BINARY
#define O_BINARY	0
#endif

using namespace Firebird;
using namespace Jrd;
using namespace Replication;
using namespace ReplicationUtils;

#if !defined(WIN_NT) && !defined(LINUX)
#pragma FB_COMPILER_MESSAGE("Add support for your platform!")
#endif

namespace
{
	const char* REPL_FILE = "fb_repl_%s";
	const ULONG SHMEM_VERSION = 1;

	const int MAX_LOG_FILES = 1000;

	const int NO_SPACE_TIMEOUT = 10; // seconds
	const int NO_SPACE_RETRIES = 6; // up to one minute

	const ULONG COPY_BLOCK_SIZE = 64 * 1024; // 64 KB

	const char* LOGFILE_PATTERN = "%s.log-%03d";
	const char* ARCHFILE_PATTERN = "%s.arch-%09d";

	const char* LOGFILENAME_WILDCARD = "$(logfilename)";
	const char* LOGPATHNAME_WILDCARD = "$(logpathname)";
	const char* ARCHFILENAME_WILDCARD = "$(archfilename)";
	const char* ARCHPATHNAME_WILDCARD = "$(archpathname)";

	const ULONG STATE_MAPPING_SIZE = 64 * 1024;	// 64 KB

	LogHeader g_dummyHeader;

	static THREAD_ENTRY_DECLARE archiver_thread(THREAD_ENTRY_PARAM arg)
	{
		ReplicationLog* const replLog = static_cast<ReplicationLog*>(arg);
		replLog->bgArchiver();
		return 0;
	}
}


// ReplicationLog::Segment class

ReplicationLog::Segment::Segment(MemoryPool& pool, const PathName& filename, int handle)
	: m_filename(pool, filename), m_handle(handle)
{
	struct stat stats;
	if (fstat(m_handle, &stats) < 0 || stats.st_size < (int) sizeof(LogHeader))
	{
		m_header = &g_dummyHeader;
		return;
	}

#ifdef WIN_NT
	m_mapping = CreateFileMapping((HANDLE) _get_osfhandle(handle), NULL, PAGE_READWRITE,
								  0, sizeof(LogHeader), NULL);
										
	if (m_mapping == INVALID_HANDLE_VALUE)
		raiseError("Log file %s mapping failed (error %d)", m_filename.c_str(), ERRNO);
		
	void* address = MapViewOfFile(m_mapping, FILE_MAP_READ | FILE_MAP_WRITE,
								  0, 0, sizeof(LogHeader));

	if (!address)
		raiseError("Log file %s mapping failed (error %d)", m_filename.c_str(), ERRNO);
#else
	void* address = mmap(NULL, sizeof(LogHeader), PROT_READ | PROT_WRITE, MAP_SHARED, handle, 0);

	if (address == MAP_FAILED)
		raiseError("Log file %s mapping failed (error %d)", m_filename.c_str(), ERRNO);
#endif
	m_header = (LogHeader*) address;
}

ReplicationLog::Segment::~Segment()
{
	if (m_header != &g_dummyHeader)
	{
#ifdef WIN_NT
		UnmapViewOfFile(m_header);
		CloseHandle(m_mapping);
#else
		munmap(m_header, sizeof(LogHeader));
#endif
	}

	::close(m_handle);
}

ReplicationLog::Segment* ReplicationLog::Segment::create(MemoryPool& pool,
														 const PathName& prefix,
														 const PathName& dir)
{
	PathName filename;

	for (int i = 0; i < MAX_LOG_FILES; i++)
	{
		filename.printf(LOGFILE_PATTERN, prefix.c_str(), i);
		filename = dir + filename;
#ifdef WIN_NT
#define S_IRGRP 0
#define S_IWGRP 0
#endif	
		const int handle =
			::open(filename.c_str(),
				   O_CREAT | O_RDWR | O_EXCL | O_BINARY | SYNC,
				   S_IREAD | S_IWRITE | S_IRGRP | S_IWGRP);
		if (handle > 0)
		{
#ifndef WIN_NT
			const char* FIREBIRD = "firebird";
			const uid_t uid = (geteuid() == 0) ? os_utils::get_user_id(FIREBIRD) : -1;
			const gid_t gid = os_utils::get_user_group_id(FIREBIRD);
			while (fchown(handle, uid, gid) < 0 && SYSCALL_INTERRUPTED(errno))
				;
			const mode_t mode = 0664;
			while (fchmod(handle, mode) < 0 && SYSCALL_INTERRUPTED(errno))
				;
#endif
			if (::write(handle, &g_dummyHeader, sizeof(LogHeader)) != sizeof(LogHeader))
				raiseError("Log file %s write failed (error %d)", filename.c_str(), ERRNO);

			return FB_NEW_POOL(pool) Segment(pool, filename, handle);
		}
	}

	raiseError("Too many log files in directory %s", dir.c_str());
	return NULL;
}

ReplicationLog::Segment* ReplicationLog::Segment::open(MemoryPool& pool, const PathName& filename)
{
	const int handle = ::open(filename.c_str(), O_RDWR | O_BINARY | SYNC, S_IREAD | S_IWRITE);

	if (handle < 0)
		raiseError("Log file %s open failed (error %d)", filename.c_str(), ERRNO);

	return FB_NEW_POOL(pool) Segment(pool, filename, handle);
}

void ReplicationLog::Segment::init(FB_UINT64 sequence, const Guid& guid)
{
	fb_assert(sizeof(LOG_SIGNATURE) == sizeof(m_header->hdr_signature));
	strcpy(m_header->hdr_signature, LOG_SIGNATURE);
	m_header->hdr_version = LOG_CURRENT_VERSION;
	m_header->hdr_state = LOG_STATE_USED;
	memcpy(&m_header->hdr_guid, &guid, sizeof(Guid));
	m_header->hdr_protocol = PROTOCOL_VERSION2;
	m_header->hdr_sequence = sequence;
	m_header->hdr_length = sizeof(LogHeader);

	flushHeader();
}

bool ReplicationLog::Segment::validate(const Guid& guid) const
{
	if (strcmp(m_header->hdr_signature, LOG_SIGNATURE))
		return false;

	if (m_header->hdr_version != LOG_CURRENT_VERSION)
		return false;

	if (m_header->hdr_state != LOG_STATE_FREE &&
		m_header->hdr_state != LOG_STATE_USED &&
		m_header->hdr_state != LOG_STATE_FULL &&
		m_header->hdr_state != LOG_STATE_ARCH)
	{
		return false;
	}

	if (memcmp(&m_header->hdr_guid, &guid, sizeof(Guid)))
		return false;

	if (m_header->hdr_protocol != PROTOCOL_VERSION2)
		return false;

	return true;
}

void ReplicationLog::Segment::append(ULONG length, const UCHAR* data, const TimeStamp& ts)
{
	fb_assert(m_header->hdr_state == LOG_STATE_USED);
	fb_assert(length);
	fb_assert(data[0] != eInitialize);

	const ULONG currentLength = m_header->hdr_length;

	if (::lseek(m_handle, currentLength, SEEK_SET) != currentLength)
		raiseError("Log file %s seek failed (error %d)", m_filename.c_str(), ERRNO);

	if (::write(m_handle, &ts.value(), sizeof(ISC_TIMESTAMP)) != sizeof(ISC_TIMESTAMP))
		raiseError("Log file %s write failed (error %d)", m_filename.c_str(), ERRNO);

	if (::write(m_handle, &length, sizeof(ULONG)) != sizeof(ULONG))
		raiseError("Log file %s write failed (error %d)", m_filename.c_str(), ERRNO);

	if (::write(m_handle, data, length) != length)
		raiseError("Log file %s write failed (error %d)", m_filename.c_str(), ERRNO);

	m_header->hdr_length += length + sizeof(ULONG) + sizeof(ISC_TIMESTAMP);

	flushHeader();
}

void ReplicationLog::Segment::copyTo(const PathName& filename) const
{
	const ULONG totalLength = m_header->hdr_length;
	fb_assert(totalLength > sizeof(LogHeader));

	const int srcHandle = ::open(m_filename.c_str(), O_RDONLY | O_BINARY, S_IREAD);

	if (srcHandle == -1)
		raiseError("Log file %s copy failed (error %d)", m_filename.c_str(), ERRNO);

	AutoFile srcFile(srcHandle);

	const int dstHandle =
		::open(filename.c_str(), O_CREAT | O_EXCL | O_WRONLY | O_BINARY, S_IREAD | S_IWRITE);

	if (dstHandle == -1)
		raiseError("Log file %s copy failed (error %d)", filename.c_str(), ERRNO);

	AutoFile dstFile(dstHandle);

	Vector<UCHAR, COPY_BLOCK_SIZE> buffer;
	UCHAR* const data = buffer.begin();

	for (ULONG offset = 0; offset < totalLength; offset += COPY_BLOCK_SIZE)
	{
		const ULONG remaining = totalLength - offset;
		const ULONG length = MIN(remaining, COPY_BLOCK_SIZE);

		if (::read(srcFile, data, length) != length)
			raiseError("Log file %s read failed (error %d)", m_filename.c_str(), ERRNO);

		if (::write(dstFile, data, length) != length)
			raiseError("Log file %s write failed (error %d)", filename.c_str(), ERRNO);
	}
}

void ReplicationLog::Segment::setState(ULONG state)
{
	m_header->hdr_state = state;
	flushHeader();
}

void ReplicationLog::Segment::truncate()
{
#ifdef WIN_NT
	::chsize(m_handle, m_header->hdr_length);
#else
	::ftruncate(m_handle, m_header->hdr_length);
#endif
}

void ReplicationLog::Segment::flushHeader()
{
	if (m_header->hdr_state == LOG_STATE_FREE)
		m_header->hdr_length = 0;

#ifdef WIN_NT
	FlushViewOfFile(m_header, 0);
#else
	msync(m_header, sizeof(LogHeader), MS_SYNC);
#endif
}

PathName ReplicationLog::Segment::getFileName() const
{
	PathName directory, filename;
	PathUtils::splitLastComponent(directory, filename, m_filename);

	return filename;
}

PathName ReplicationLog::Segment::getArchiveName() const
{
	const PathName filename = getFileName();

	const size_t pos = filename.find_last_of('.');
	fb_assert(pos != string::npos);

	PathName archname;
	archname.printf(ARCHFILE_PATTERN, filename.substr(0, pos).c_str(), (ULONG) m_header->hdr_sequence);
	return archname;
}


// ReplicationLog class

ReplicationLog::ReplicationLog(MemoryPool& pool, ReplicationConfig* config,
							   const PathName& database, const Guid& guid,
							   bool async)
	: PermanentStorage(pool), ErrorLogger(pool, database, MASTER),
	  m_config(config), m_segments(pool), m_shutdown(false), m_async(async)
{
	memcpy(&m_guid, &guid, sizeof(Guid));

	char buff[GUID_BUFF_SIZE];
	GuidToString(buff, &guid);
	const string guidStr(buff);

	PathName filename;
	filename.printf(REPL_FILE, guidStr.c_str());

	FB_NEW_POOL(pool) SharedMemory<State>(filename.c_str(), STATE_MAPPING_SIZE, this);

	{ // scope
		LockGuard guard(this);
/*
 *	dimitr: If the server crashes while archiving, segments will remain in the ARCH state
 *			forever. The code underneath allows to recover their state and retry archiving them.
 *			However, it may cause errors if the crash happened after the archive destination has
 *			already been created -- the "already exists" case -- and pollute the log with errors
 *			while infinitely (and unsuccessfully) retrying to re-archive the segment. Sigh.
 *
		State* const state = m_state->getHeader();

		if (!state->pidUpper)
		{
			fb_assert(!state->pidLower);

			for (Segment** iter = m_segments.begin(); iter != m_segments.end(); ++iter)
			{
				Segment* const segment = *iter;

				if (segment->getState() == LOG_STATE_ARCH)
					segment->setState(LOG_STATE_FULL);
			}
		}
*/
		linkSelf();
	}

	if (m_async)
	{
		Thread::start(archiver_thread, this, THREAD_high, 0);
		m_startupSemaphore.enter();
		m_workingSemaphore.release();
	}
}

ReplicationLog::~ReplicationLog()
{
	m_shutdown = true;

	if (m_async)
	{
		m_workingSemaphore.release();
		m_cleanupSemaphore.enter();

		try
		{
			LockGuard guard(this);

			if (unlinkSelf())
				switchActiveSegment();
		}
		catch (const Exception&)
		{}	// no-op
	}

	clearSegments();

	if (m_state.hasData() && m_state->getHeader())
		delete m_state.release();
}

void ReplicationLog::lockState()
{
	bool blockage = false;

	if (!m_state->mutexLockCond())
	{
		blockage = true;
		m_state->mutexLock();
	}

	try
	{
		State* const state = m_state->getHeader();

		state->lockAcquires++;
		if (blockage)
			state->lockBlocks++;

		if (m_segments.isEmpty() || state->segmentCount > (ULONG) m_segments.getCount())
			initSegments();
	}
	catch (Exception&)
	{
		unlockState();
		throw;
	}
}

void ReplicationLog::unlockState()
{
	m_state->mutexUnlock();
}

void ReplicationLog::linkSelf()
{
	const int process_id = getpid();

	State* const state = m_state->getHeader();

	if (state->pidLower == state->pidUpper)
	{
		bool found = false;

		for (ULONG i = 0; i < state->pidUpper; i++)
		{
			if (state->pids[i] == process_id ||
				!ISC_check_process_existence(state->pids[i]))
			{
				state->pids[i] = process_id;
				found = true;
				break;
			}
		}

		if (!found)
		{
			const ULONG capacity =
				(STATE_MAPPING_SIZE - offsetof(State, pids)) / sizeof(int);

			if (state->pidUpper == capacity)
				status_exception::raise(Arg::Gds(isc_imp_exc));

			state->pids[state->pidUpper++] = process_id;
			state->pidLower = state->pidUpper;
		}
	}
	else
	{
		fb_assert(!state->pids[state->pidLower]);
		state->pids[state->pidLower++] = process_id;

		while (state->pidLower < state->pidUpper)
		{
			if (!state->pids[state->pidLower++])
				break;
		}
	}
}

bool ReplicationLog::unlinkSelf()
{
	const int process_id = getpid();

	State* const state = m_state->getHeader();

	fb_assert(state->pidUpper);

	ULONG empty_slots = 0;

	for (ULONG i = 0; i < state->pidUpper; i++)
	{
		if (state->pids[i])
		{
			if (state->pids[i] != process_id &&
				ISC_check_process_existence(state->pids[i]))
			{
				empty_slots = 0;
				continue;
			}

			state->pids[i] = 0;
			state->pidLower = MIN(state->pidLower, i);
		}

		empty_slots++;
	}

	state->pidUpper -= empty_slots;

	return (state->pidUpper == 0);
}

bool ReplicationLog::initialize(SharedMemoryBase* shmem, bool init)
{
	m_state.reset(reinterpret_cast<SharedMemory<State>*>(shmem));

	if (init)
	{
		State* const state = m_state->getHeader();
		memset(state, 0, sizeof(State));

		state->init(SharedMemoryBase::SRAM_REPL_STATE, SHMEM_VERSION);

		state->timestamp = time(NULL);
	}

	return true;
}

void ReplicationLog::mutexBug(int osErrorCode, const char* /*text*/)
{
	raiseError("Shared memory locking failed (error %d)", osErrorCode);
}

void ReplicationLog::write(ULONG length, const UCHAR* data)
{
	const TimeStamp ts = TimeStamp::getCurrentTimeStamp();

	LockGuard guard(this);

	Segment* segment = getSegment(length);

	for (int i = 0; i < NO_SPACE_RETRIES && !segment; i++)
	{
		logMessage("Out of available space in replication log segments, waiting for archiving...");

		{	// scope
			LockCheckout checkout(this);
			Thread::sleep(NO_SPACE_TIMEOUT);
		}

		segment = getSegment(length);
	}

	if (!segment)
		raiseError("Out of available space in replication log segments");

	State* const state = m_state->getHeader();
	if (segment->getLength() == sizeof(LogHeader))
		state->timestamp = time(NULL);

	segment->append(length, data, ts);
}

bool ReplicationLog::archiveExecute(const Segment* segment)
{
	LockCheckout checkout(this);

	const PathName logfilename = segment->getFileName();
	const PathName logpathname = m_config->logDirectory + logfilename;

	const PathName archfilename = segment->getArchiveName();
	const PathName archpathname = m_config->archiveDirectory + archfilename;

	string archiveCommand = m_config->archiveCommand;

	if (archiveCommand.hasData())
	{
		size_t pos;
		while ( (pos = archiveCommand.find(LOGPATHNAME_WILDCARD)) != string::npos)
			archiveCommand.replace(pos, strlen(LOGPATHNAME_WILDCARD), logpathname);

		while ( (pos = archiveCommand.find(LOGFILENAME_WILDCARD)) != string::npos)
			archiveCommand.replace(pos, strlen(LOGFILENAME_WILDCARD), logfilename);

		while ( (pos = archiveCommand.find(ARCHFILENAME_WILDCARD)) != string::npos)
			archiveCommand.replace(pos, strlen(ARCHFILENAME_WILDCARD), archfilename);

		while ( (pos = archiveCommand.find(ARCHPATHNAME_WILDCARD)) != string::npos)
			archiveCommand.replace(pos, strlen(ARCHPATHNAME_WILDCARD), archpathname);

		fb_assert(archiveCommand.hasData());
		const int res = executeShell(archiveCommand);

		if (res)
		{
			string errorMsg;

			if (res < 0)
			{
				errorMsg.printf("Error executing log archive command (error %d): %s",
								ERRNO, archiveCommand.c_str());
			}
			else
			{
				errorMsg.printf("Unexpected result (%d) while executing log archive command: %s",
								res, archiveCommand.c_str());
			}

			logMessage(errorMsg);
			return false;
		}
	}
	else
	{
		try
		{
			segment->copyTo(archpathname);
		}
		catch (const Exception& ex)
		{
			string errorMsg;
			errorMsg.printf("Error while copying log segment: %s", ex.what());

			logMessage(errorMsg);
			return false;
		}
	}

	return true;
}

bool ReplicationLog::archiveSegment(Segment* segment)
{
	if (m_config->archiveCommand.hasData() || m_config->archiveDirectory.hasData())
	{
		if (m_config->archiveCommand.hasData())
			segment->truncate();

		segment->setState(LOG_STATE_ARCH);
		segment->addRef();

		const bool success = archiveExecute(segment);

		fb_assert(segment->getState() == LOG_STATE_ARCH);
		segment->setState(success ? LOG_STATE_FREE : LOG_STATE_FULL);
		segment->release();

		return success;
	}

	return false;
}

void ReplicationLog::switchActiveSegment()
{
	Segment* activeSegment = NULL;
	ULONG full_segments = 0;

	for (Segment** iter = m_segments.begin(); iter != m_segments.end(); ++iter)
	{
		Segment* const segment = *iter;

		const ULONG state = segment->getState();

		if (state == LOG_STATE_USED)
			activeSegment = segment;
		else if (state == LOG_STATE_FULL)
			full_segments++;
	}

	if (activeSegment)
	{
		activeSegment->setState(LOG_STATE_FULL);

		if (!full_segments)
			archiveSegment(activeSegment);
	}
}

void ReplicationLog::bgArchiver()
{
	fb_assert(m_async);

	m_startupSemaphore.release();

	try
	{
		while (!m_shutdown)
		{
			LockGuard guard(this);

			State* const state = m_state->getHeader();

			for (Segment** iter = m_segments.begin(); iter != m_segments.end(); ++iter)
			{
				Segment* const segment = *iter;

				if (segment->getState() == LOG_STATE_USED)
				{
					if (segment->getLength() > sizeof(LogHeader) && m_config->archiveTimeout)
					{
						const size_t delta_timestamp = time(NULL) - state->timestamp;

						if (delta_timestamp > m_config->archiveTimeout)
							segment->setState(LOG_STATE_FULL);
					}

					break;
				}
			}

			while (!m_shutdown)
			{
				bool restart = false;

				for (Segment** iter = m_segments.begin(); iter != m_segments.end(); ++iter)
				{
					Segment* const segment = *iter;

					if (segment->getState() == LOG_STATE_FULL)
					{
						if (archiveSegment(segment))
						{
							restart = true;
							break;
						}
					}
				}

				if (!restart)
					break;
			}

			guard.release();

			m_workingSemaphore.tryEnter(1);
		}
	}
	catch (const Firebird::Exception& ex)
	{
		iscLogException("Error in replication log thread", ex);
	}

	m_cleanupSemaphore.release();
}

void ReplicationLog::initSegments()
{
	clearSegments();

	FB_UINT64 maxSequence = 0;

	for (DirectoryWalker iter(m_config->logDirectory); iter.hasData(); ++iter)
	{
		const PathName filename = *iter;

		AutoPtr<Segment> segment(Segment::open(getPool(), filename));

		if (!validateSegment(segment))
			continue;

		maxSequence = MAX(segment->getSequence(), maxSequence);

		segment->addRef();
		m_segments.add(segment.release());
	}

	State* const state = m_state->getHeader();
	state->segmentCount = (ULONG) m_segments.getCount();
	state->sequence = maxSequence;
}

void ReplicationLog::clearSegments()
{
	while (m_segments.hasData())
		m_segments.pop()->release();
}

ReplicationLog::Segment* ReplicationLog::getSegment(ULONG length)
{
	Segment* activeSegment = NULL;
	Segment* freeSegment = NULL;

	for (Segment** iter = m_segments.begin(); iter != m_segments.end(); ++iter)
	{
		Segment* const segment = *iter;

		const ULONG state = segment->getState();

		if (state == LOG_STATE_USED)
		{
			if (activeSegment)
				raiseError("Multiple active replication log segments found");

			activeSegment = segment;
		}
		else if (state == LOG_STATE_FREE)
		{
			if (!freeSegment)
				freeSegment = segment;
		}
	}

	if (activeSegment)
	{
		if (activeSegment->getLength() + length > m_config->segmentSize)
		{
			activeSegment->setState(LOG_STATE_FULL);
			activeSegment = NULL;
			m_workingSemaphore.release();
		}
		else if (activeSegment->getLength() > sizeof(LogHeader) && m_config->archiveTimeout)
		{
			State* const state = m_state->getHeader();
			const size_t delta_timestamp = time(NULL) - state->timestamp;

			if (delta_timestamp > m_config->archiveTimeout)
			{
				activeSegment->setState(LOG_STATE_FULL);
				activeSegment = NULL;
				m_workingSemaphore.release();
			}
		}
	}

	if (activeSegment)
		return activeSegment;

	if (freeSegment)
	{
		initSegment(freeSegment);
		return freeSegment;
	}

	if (!m_config->segmentCount || m_segments.getCount() < m_config->segmentCount)
		return createSegment();

	return NULL;
}

ReplicationLog::Segment* ReplicationLog::createSegment()
{
	AutoPtr<Segment> segment(Segment::create(getPool(),
							 m_config->logFilePrefix,
							 m_config->logDirectory));

	initSegment(segment);

	State* const state = m_state->getHeader();
	state->segmentCount++;

	segment->addRef();
	m_segments.add(segment);

	return segment.release();
}
